VERIFICATION_QUERY = """
query VerificationQuery {
    version
}
"""

ASSET_NODES_QUERY = """
query AssetNodeQuery {
    assetNodes {
        id
        assetKey {
            path
        }
        metadataEntries {
            ... on TextMetadataEntry {
                label
                text
            }
            ... on JsonMetadataEntry {
                label
                jsonString
            }
            __typename
        }
        jobs {
            id
            name
            repository {
                id
                name
                location {
                    id
                    name
                }
            }
        }
        isPartitioned
        partitionDefinition {
          type
          name
          fmt
        }
        partitionKeys
    }
}
"""

TRIGGER_ASSETS_MUTATION = """
mutation LaunchAssetsExecution($executionParams: ExecutionParams!) {
  launchPipelineExecution(executionParams: $executionParams) {
    ... on LaunchRunSuccess {
      run {
        id
        pipelineName
        __typename
      }
      __typename
    }
    ... on PipelineNotFoundError {
      message
      __typename
    }
    ... on InvalidSubsetError {
      message
      __typename
    }
    ... on RunConfigValidationInvalid {
      errors {
        message
        __typename
      }
      __typename
    }
    ...PythonErrorFragment
    __typename
  }
}

fragment PythonErrorFragment on PythonError {
  message
  stack
  errorChain {
    ...PythonErrorChain
    __typename
  }
  __typename
}

fragment PythonErrorChain on ErrorChainLink {
  isExplicitLink
  error {
    message
    stack
    __typename
  }
  __typename
}
"""

RUNS_QUERY = """
query RunQuery($runId: ID!) {
	runOrError(runId: $runId) {
		__typename
		...PythonErrorFragment
		...NotFoundFragment
		... on Run {
			id
			status
      tags {
        key
        value
      }
			__typename
		}
	}
}
fragment NotFoundFragment on RunNotFoundError {
	__typename
	message
}
fragment PythonErrorFragment on PythonError {
	__typename
	message
	stack
	causes {
		message
		stack
		__typename
	}
}
"""

RUNS_BY_TAG_QUERY = """
query RunsByTagQuery($filter: RunsFilter!) {
  runsOrError(filter: $filter) {
    ... on Runs {
      results {
        id
        status
        tags {
          key
          value
        }
      }
    }
  }
}
"""



from collections import defaultdict
from collections.abc import Callable, Iterable, Mapping, Sequence
from dataclasses import dataclass
from typing import Optional, Union, cast

from dagster import AssetsDefinition, AssetSpec, Definitions
from dagster._annotations import beta
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.assets.definition.asset_spec import map_asset_specs
from dagster._core.definitions.definitions_load_context import StateBackedDefinitionsLoader
from dagster._core.definitions.external_asset import external_asset_from_spec
from dagster._core.definitions.sensor_definition import DefaultSensorStatus

from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.job_builder import construct_dag_jobs
from dagster_airlift.core.monitoring_job.builder import build_airflow_monitoring_defs
from dagster_airlift.core.sensor.event_translation import (
    DagsterEventTransformerFn,
    default_event_transformer,
)
from dagster_airlift.core.sensor.sensor_builder import (
    DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS,
    build_airflow_polling_sensor,
)
from dagster_airlift.core.serialization.compute import DagSelectorFn, compute_serialized_data
from dagster_airlift.core.serialization.defs_construction import (
    construct_dag_assets_defs,
    get_airflow_data_to_spec_mapper,
)
from dagster_airlift.core.serialization.serialized_data import (
    DagInfo,
    SerializedAirflowDefinitionsData,
)
from dagster_airlift.core.utils import (
    MappedAsset,
    dag_handles_for_spec,
    get_metadata_key,
    is_dag_mapped_asset_spec,
    is_task_mapped_asset_spec,
    spec_iterator,
    task_handles_for_spec,
    type_narrow_defs_assets,
)


@dataclass
class AirflowInstanceDefsLoader(StateBackedDefinitionsLoader[SerializedAirflowDefinitionsData]):
    airflow_instance: AirflowInstance
    retrieval_filter: AirflowFilter
    mapped_assets: Sequence[MappedAsset]
    source_code_retrieval_enabled: Optional[bool]
    sensor_minimum_interval_seconds: int = DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS
    dag_selector_fn: Optional[DagSelectorFn] = None

    @property
    def defs_key(self) -> str:
        return get_metadata_key(self.airflow_instance.name)

    def fetch_state(self) -> SerializedAirflowDefinitionsData:
        return compute_serialized_data(
            airflow_instance=self.airflow_instance,
            mapped_assets=self.mapped_assets,
            dag_selector_fn=self.dag_selector_fn,
            automapping_enabled=False,
            source_code_retrieval_enabled=self.source_code_retrieval_enabled,
            retrieval_filter=self.retrieval_filter,
        )

    def defs_from_state(  # pyright: ignore[reportIncompatibleMethodOverride]
        self, serialized_airflow_data: SerializedAirflowDefinitionsData
    ) -> Definitions:
        raise Exception(
            "We use get_or_fetch_state() to build definitions, and leave it up to the callsite how it is used."
        )


@beta
def build_defs_from_airflow_instance(
    *,
    airflow_instance: AirflowInstance,
    defs: Optional[Definitions] = None,
    sensor_minimum_interval_seconds: int = DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS,
    event_transformer_fn: DagsterEventTransformerFn = default_event_transformer,
    dag_selector_fn: Optional[Callable[[DagInfo], bool]] = None,
    source_code_retrieval_enabled: Optional[bool] = None,
    default_sensor_status: Optional[DefaultSensorStatus] = None,
    retrieval_filter: Optional[AirflowFilter] = None,
) -> Definitions:
    """Builds a :py:class:`dagster.Definitions` object from an Airflow instance.

    For every DAG in the Airflow instance, this function will create a Dagster asset for the DAG
    with an asset key instance_name/dag/dag_id. It will also create a sensor that polls the Airflow
    instance for DAG runs and emits Dagster events for each successful run.

    An optional `defs` argument can be provided, where the user can pass in a :py:class:`dagster.Definitions`
    object containing assets which are mapped to Airflow DAGs and tasks. These assets will be enriched with
    metadata from the Airflow instance, and placed upstream of the automatically generated DAG assets.

    An optional `event_transformer_fn` can be provided, which allows the user to modify the Dagster events
    produced by the sensor. The function takes the Dagster events produced by the sensor and returns a sequence
    of Dagster events.

    An optional `dag_selector_fn` can be provided, which allows the user to filter which DAGs assets are created for.
    The function takes a :py:class:`dagster_airlift.core.serialization.serialized_data.DagInfo` object and returns a
    boolean indicating whether the DAG should be included.

    Args:
        airflow_instance (AirflowInstance): The Airflow instance to build assets and the sensor from.
        defs: Optional[Definitions]: A :py:class:`dagster.Definitions` object containing assets that are
            mapped to Airflow DAGs and tasks.
        sensor_minimum_interval_seconds (int): The minimum interval in seconds between sensor runs.
        event_transformer_fn (DagsterEventTransformerFn): A function that allows for modifying the Dagster events
            produced by the sensor.
        dag_selector_fn (Optional[Callable[[DagInfo], bool]]): A function that allows for filtering which DAGs assets are created for.
        source_code_retrieval_enabled (Optional[bool]): Whether to retrieve source code for the Airflow DAGs. By default, source code is retrieved when the number of DAGs is under 50 for performance reasons. This setting overrides the default behavior.
        default_sensor_status (Optional[DefaultSensorStatus]): The default status for the sensor. By default, the sensor will be enabled.

    Returns:
        Definitions: A :py:class:`dagster.Definitions` object containing the assets and sensor.

    Examples:
        Building a :py:class:`dagster.Definitions` object from an Airflow instance.

        .. code-block:: python

            from dagster_airlift.core import (
                AirflowInstance,
                AirflowBasicAuthBackend,
                build_defs_from_airflow_instance,
            )

            from .constants import AIRFLOW_BASE_URL, AIRFLOW_INSTANCE_NAME, PASSWORD, USERNAME

            airflow_instance = AirflowInstance(
                auth_backend=AirflowBasicAuthBackend(
                    webserver_url=AIRFLOW_BASE_URL, username=USERNAME, password=PASSWORD
                ),
                name=AIRFLOW_INSTANCE_NAME,
            )


            defs = build_defs_from_airflow_instance(airflow_instance=airflow_instance)

        Providing task-mapped assets to the function.

        .. code-block:: python

            from dagster import Definitions
            from dagster_airlift.core import (
                AirflowInstance,
                AirflowBasicAuthBackend,
                assets_with_task_mappings,
                build_defs_from_airflow_instance,
            )
            ...


            defs = build_defs_from_airflow_instance(
                airflow_instance=airflow_instance, # same as above
                defs=Definitions(
                    assets=assets_with_task_mappings(
                        dag_id="rebuild_iris_models",
                        task_mappings={
                            "my_task": [AssetSpec("my_first_asset"), AssetSpec("my_second_asset")],
                        },
                    ),
                ),
            )

        Providing a custom event transformer function.

        .. code-block:: python

            from typing import Sequence
            from dagster import Definitions, SensorEvaluationContext
            from dagster_airlift.core import (
                AirflowInstance,
                AirflowBasicAuthBackend,
                AssetEvent,
                assets_with_task_mappings,
                build_defs_from_airflow_instance,
                AirflowDefinitionsData,
            )
            ...

            def add_tags_to_events(
                context: SensorEvaluationContext,
                defs_data: AirflowDefinitionsData,
                events: Sequence[AssetEvent]
            ) -> Sequence[AssetEvent]:
                altered_events = []
                for event in events:
                    altered_events.append(event._replace(tags={"my_tag": "my_value"}))
                return altered_events

            defs = build_defs_from_airflow_instance(
                airflow_instance=airflow_instance, # same as above
                event_transformer_fn=add_tags_to_events,
            )

        Filtering which DAGs assets are created for.

        .. code-block:: python

            from dagster import Definitions
            from dagster_airlift.core import (
                AirflowInstance,
                AirflowBasicAuthBackend,
                AssetEvent,
                assets_with_task_mappings,
                build_defs_from_airflow_instance,
                DagInfo,
            )
            ...

            def only_include_dag(dag_info: DagInfo) -> bool:
                return dag_info.dag_id == "my_dag_id"

            defs = build_defs_from_airflow_instance(
                airflow_instance=airflow_instance, # same as above
                dag_selector_fn=only_include_dag,
            )

    """
    defs = defs or Definitions()
    mapped_assets = type_narrow_defs_assets(defs)
    serialized_airflow_data = AirflowInstanceDefsLoader(
        airflow_instance=airflow_instance,
        mapped_assets=mapped_assets,
        dag_selector_fn=dag_selector_fn,
        source_code_retrieval_enabled=source_code_retrieval_enabled,
        retrieval_filter=retrieval_filter or AirflowFilter(),
    ).get_or_fetch_state()
    assets_to_apply_airflow_data = [
        *mapped_assets,
        *construct_dataset_specs(serialized_airflow_data),
    ]
    mapped_and_constructed_assets = [
        *_apply_airflow_data_to_specs(assets_to_apply_airflow_data, serialized_airflow_data),
        *construct_dag_assets_defs(serialized_airflow_data),
    ]
    fully_resolved_assets_definitions = [
        external_asset_from_spec(asset)
        if isinstance(asset, AssetSpec)
        else cast("AssetsDefinition", asset)
        for asset in mapped_and_constructed_assets
    ]
    defs_with_airflow_assets = replace_assets_in_defs(
        defs=defs, assets=fully_resolved_assets_definitions
    )

    return Definitions.merge_unbound_defs(
        defs_with_airflow_assets,
        Definitions(
            sensors=[
                build_airflow_polling_sensor(
                    mapped_assets=mapped_and_constructed_assets,
                    airflow_instance=airflow_instance,
                    minimum_interval_seconds=sensor_minimum_interval_seconds,
                    event_transformer_fn=event_transformer_fn,
                    default_sensor_status=default_sensor_status,
                )
            ]
        ),
    )


def _apply_airflow_data_to_specs(
    assets: Sequence[MappedAsset],
    serialized_data: SerializedAirflowDefinitionsData,
) -> Sequence[MappedAsset]:
    """Apply asset spec transformations to the assets."""
    return cast(
        "Sequence[MappedAsset]",
        map_asset_specs(
            func=get_airflow_data_to_spec_mapper(serialized_data),
            iterable=assets,
        ),
    )


def replace_assets_in_defs(
    defs: Definitions, assets: Iterable[Union[AssetSpec, AssetsDefinition]]
) -> Definitions:
    return Definitions(
        assets=list(assets),
        asset_checks=defs.asset_checks,
        sensors=defs.sensors,
        schedules=defs.schedules,
        jobs=defs.jobs,
        executor=defs.executor,
        loggers=defs.loggers,
        resources=defs.resources,
    )


def enrich_airflow_mapped_assets(
    mapped_assets: Sequence[MappedAsset],
    airflow_instance: AirflowInstance,
    source_code_retrieval_enabled: Optional[bool],
    retrieval_filter: Optional[AirflowFilter] = None,
) -> Sequence[MappedAsset]:
    """Enrich Airflow-mapped assets with metadata from the provided :py:class:`AirflowInstance`."""
    serialized_data = AirflowInstanceDefsLoader(
        airflow_instance=airflow_instance,
        mapped_assets=mapped_assets,
        source_code_retrieval_enabled=source_code_retrieval_enabled,
        retrieval_filter=retrieval_filter or AirflowFilter(),
    ).get_or_fetch_state()
    return list(_apply_airflow_data_to_specs(mapped_assets, serialized_data))


@beta
def load_airflow_dag_asset_specs(
    airflow_instance: AirflowInstance,
    mapped_assets: Optional[Sequence[MappedAsset]] = None,
    dag_selector_fn: Optional[Callable[[DagInfo], bool]] = None,
    source_code_retrieval_enabled: Optional[bool] = None,
    retrieval_filter: Optional[AirflowFilter] = None,
) -> Sequence[AssetSpec]:
    """Load asset specs for Airflow DAGs from the provided :py:class:`AirflowInstance`, and link upstreams from mapped assets."""
    serialized_data = AirflowInstanceDefsLoader(
        airflow_instance=airflow_instance,
        mapped_assets=mapped_assets or [],
        dag_selector_fn=dag_selector_fn,
        source_code_retrieval_enabled=source_code_retrieval_enabled,
        retrieval_filter=retrieval_filter or AirflowFilter(),
    ).get_or_fetch_state()
    return list(spec_iterator(construct_dag_assets_defs(serialized_data)))


def uri_to_asset_key(uri: str) -> AssetKey:
    last_path_segment = uri.split("/")[-1]
    with_ext_removed = last_path_segment.split(".")[0]
    return AssetKey(with_ext_removed)


def construct_dataset_specs(
    serialized_data: SerializedAirflowDefinitionsData,
) -> Sequence[AssetSpec]:
    """Construct dataset definitions from the serialized Airflow data."""
    from dagster_airlift.core.multiple_tasks import assets_with_multiple_task_mappings

    return cast(
        "Sequence[AssetSpec]",
        [
            assets_with_multiple_task_mappings(
                task_handles=[
                    {"dag_id": t.dag_id, "task_id": t.task_id} for t in dataset.producing_tasks
                ],
                assets=[
                    AssetSpec(
                        key=uri_to_asset_key(dataset.uri),
                        metadata=dataset.extra,
                        deps=[
                            uri_to_asset_key(upstream_uri)
                            for upstream_uri in serialized_data.upstream_datasets_by_uri.get(
                                dataset.uri, set()
                            )
                        ],
                    )
                ],
            )[0]
            for dataset in serialized_data.datasets
        ],
    )


def _get_dag_to_spec_mapping(
    mapped_assets: Sequence[Union[AssetSpec, AssetsDefinition]],
) -> Mapping[str, Sequence[AssetSpec]]:
    res = defaultdict(list)
    for spec in spec_iterator(mapped_assets):
        if is_task_mapped_asset_spec(spec):
            for task_handle in task_handles_for_spec(spec):
                res[task_handle.dag_id].append(spec)
        elif is_dag_mapped_asset_spec(spec):
            for dag_handle in dag_handles_for_spec(spec):
                res[dag_handle.dag_id].append(spec)
    return res


def build_job_based_airflow_defs(
    *,
    airflow_instance: AirflowInstance,
    retrieval_filter: Optional[AirflowFilter] = None,
    mapped_defs: Optional[Definitions] = None,
    source_code_retrieval_enabled: Optional[bool] = None,
) -> Definitions:
    mapped_defs = mapped_defs or Definitions()
    retrieval_filter = retrieval_filter or AirflowFilter()
    mapped_assets = type_narrow_defs_assets(mapped_defs)
    serialized_airflow_data = AirflowInstanceDefsLoader(
        airflow_instance=airflow_instance,
        mapped_assets=mapped_assets,
        dag_selector_fn=None,
        source_code_retrieval_enabled=source_code_retrieval_enabled,
        retrieval_filter=retrieval_filter,
    ).get_or_fetch_state()
    assets_with_airflow_data = _apply_airflow_data_to_specs(
        [
            *mapped_assets,
            *construct_dataset_specs(serialized_airflow_data),
        ],
        serialized_airflow_data,
    )
    dag_to_spec_mapping = _get_dag_to_spec_mapping(assets_with_airflow_data)
    jobs = construct_dag_jobs(
        serialized_data=serialized_airflow_data,
        mapped_specs=dag_to_spec_mapping,
    )
    return Definitions.merge(
        replace_assets_in_defs(defs=mapped_defs, assets=assets_with_airflow_data),
        Definitions(jobs=jobs),
        build_airflow_monitoring_defs(airflow_instance=airflow_instance),
    )



from collections import defaultdict
from collections.abc import Iterable, Sequence
from functools import cached_property
from typing import TYPE_CHECKING, AbstractSet, Callable, Optional  # noqa: UP035

from dagster import AssetKey, AssetSpec
from dagster._record import record

from dagster_airlift.core.airflow_instance import AirflowInstance, DagInfo
from dagster_airlift.core.dag_asset import get_leaf_assets_for_dag
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.serialization.serialized_data import (
    DagHandle,
    Dataset,
    KeyScopedDagHandles,
    KeyScopedTaskHandles,
    SerializedAirflowDefinitionsData,
    SerializedDagData,
    TaskHandle,
    TaskInfo,
)
from dagster_airlift.core.utils import (
    dag_handles_for_spec,
    is_dag_mapped_asset_spec,
    is_task_mapped_asset_spec,
    spec_iterator,
    task_handles_for_spec,
)

if TYPE_CHECKING:
    from dagster_airlift.core.airflow_defs_data import MappedAsset


DEFAULT_MAX_NUM_DAGS_SOURCE_CODE_RETRIEVAL = 50
DagSelectorFn = Callable[[DagInfo], bool]


@record
class AirliftMetadataMappingInfo:
    asset_specs: Iterable[AssetSpec]

    @cached_property
    def mapped_task_asset_specs(self) -> list[AssetSpec]:
        return [spec for spec in self.asset_specs if is_task_mapped_asset_spec(spec)]

    @cached_property
    def mapped_dag_asset_specs(self) -> list[AssetSpec]:
        return [spec for spec in self.asset_specs if is_dag_mapped_asset_spec(spec)]

    @cached_property
    def dag_ids(self) -> set[str]:
        return set(self.all_mapped_asset_keys_by_dag_id.keys())

    @cached_property
    def task_id_map(self) -> dict[str, set[str]]:
        """Mapping of dag_id to set of task_ids in that dag. This only contains task ids mapped to assets in this object."""
        task_id_map_data = {
            dag_id: set(ta_map.keys())
            for dag_id, ta_map in self.asset_keys_by_mapped_task_id.items()
        }
        return defaultdict(set, task_id_map_data)

    @cached_property
    def all_mapped_asset_keys_by_dag_id(self) -> dict[str, set[AssetKey]]:
        """Mapping of dag_id to set of asset_keys which are materialized by that dag.

        If assets within the dag are mapped to individual tasks, all of those assets will be included in this set.
        If the dag itself is mapped to a set of assets, those assets will be included in this set.
        """
        asset_keys_in_dag_by_id = defaultdict(set)
        for dag_id, task_to_asset_map in self.asset_keys_by_mapped_task_id.items():
            for asset_keys in task_to_asset_map.values():
                asset_keys_in_dag_by_id[dag_id].update(asset_keys)
        for dag_id, asset_keys in self.asset_keys_by_mapped_dag_id.items():
            asset_keys_in_dag_by_id[dag_id].update(asset_keys)
        return defaultdict(set, asset_keys_in_dag_by_id)

    @cached_property
    def asset_keys_by_mapped_task_id(self) -> dict[str, dict[str, set[AssetKey]]]:
        """Mapping of dag_id to task_id to set of asset_keys mapped from that task."""
        asset_key_map: dict[str, dict[str, set[AssetKey]]] = defaultdict(lambda: defaultdict(set))
        for spec in self.asset_specs:
            if is_task_mapped_asset_spec(spec):
                for task_handle in task_handles_for_spec(spec):
                    asset_key_map[task_handle.dag_id][task_handle.task_id].add(spec.key)
        return asset_key_map

    @cached_property
    def asset_keys_by_mapped_dag_id(self) -> dict[str, set[AssetKey]]:
        """Mapping of dag_id to set of asset_keys mapped from that dag."""
        asset_key_map: dict[str, set[AssetKey]] = defaultdict(set)
        for spec in self.asset_specs:
            if is_dag_mapped_asset_spec(spec):
                for dag_handle in dag_handles_for_spec(spec):
                    asset_key_map[dag_handle.dag_id].add(spec.key)
        return asset_key_map

    @cached_property
    def task_handle_map(self) -> dict[AssetKey, set[TaskHandle]]:
        task_handle_map = defaultdict(set)
        for dag_id, asset_key_by_task_id in self.asset_keys_by_mapped_task_id.items():
            for task_id, asset_keys in asset_key_by_task_id.items():
                for asset_key in asset_keys:
                    task_handle_map[asset_key].add(TaskHandle(dag_id=dag_id, task_id=task_id))
        return task_handle_map

    @cached_property
    def downstream_deps(self) -> dict[AssetKey, set[AssetKey]]:
        downstreams = defaultdict(set)
        for spec in self.asset_specs:
            for dep in spec.deps:
                downstreams[dep.asset_key].add(spec.key)
        return downstreams


def build_airlift_metadata_mapping_info(
    mapped_assets: Iterable["MappedAsset"],
) -> AirliftMetadataMappingInfo:
    asset_specs = list(spec_iterator(mapped_assets))
    return AirliftMetadataMappingInfo(asset_specs=asset_specs)


@record
class FetchedAirflowData:
    dag_infos: dict[str, DagInfo]
    task_info_map: dict[str, dict[str, TaskInfo]]
    mapping_info: AirliftMetadataMappingInfo
    datasets: Sequence[Dataset]

    @cached_property
    def all_mapped_tasks(self) -> dict[AssetKey, AbstractSet[TaskHandle]]:
        return {
            spec.key: task_handles_for_spec(spec)
            for spec in self.mapping_info.mapped_task_asset_specs
        }

    @cached_property
    def all_mapped_dags(self) -> dict[AssetKey, AbstractSet[DagHandle]]:
        return {
            spec.key: dag_handles_for_spec(spec)
            for spec in self.mapping_info.mapped_dag_asset_specs
        }


def fetch_all_airflow_data(
    airflow_instance: AirflowInstance,
    mapping_info: AirliftMetadataMappingInfo,
    dag_selector_fn: Optional[DagSelectorFn],
    automapping_enabled: bool,
    retrieval_filter: AirflowFilter,
) -> FetchedAirflowData:
    dag_infos = {
        dag.dag_id: dag
        for dag in airflow_instance.list_dags(retrieval_filter=retrieval_filter)
        if dag_selector_fn is None or dag_selector_fn(dag)
    }

    # To limit the number of API calls, only fetch task infos for the dags that we absolutely have to.
    # Airflow has no batch API for fetching task infos, so we have to fetch them one dag
    # at a time.
    task_info_map = defaultdict(dict)
    for dag_id in dag_infos.keys():
        # Explicitly don't fetch task information for dags that have no mapped tasks,
        # unless automapping is enabled.
        if len(mapping_info.task_id_map[dag_id]) == 0 and not automapping_enabled:
            continue
        task_info_map[dag_id] = {
            task_info.task_id: task_info
            for task_info in airflow_instance.get_task_infos(dag_id=dag_id)
        }

    if retrieval_filter.retrieve_datasets:
        datasets = airflow_instance.get_all_datasets(
            retrieval_filter=retrieval_filter, dag_ids=list(dag_infos.keys())
        )
    else:
        datasets = []
    return FetchedAirflowData(
        dag_infos=dag_infos,
        task_info_map=task_info_map,
        mapping_info=mapping_info,
        datasets=datasets,
    )


def infer_code_retrieval_enabled(
    source_code_retrieval_enabled: Optional[bool], fetched_airflow_data: FetchedAirflowData
) -> bool:
    if source_code_retrieval_enabled is None:
        return len(fetched_airflow_data.dag_infos) < DEFAULT_MAX_NUM_DAGS_SOURCE_CODE_RETRIEVAL
    return source_code_retrieval_enabled


def compute_serialized_data(
    airflow_instance: AirflowInstance,
    mapped_assets: Iterable["MappedAsset"],
    dag_selector_fn: Optional[DagSelectorFn],
    automapping_enabled: bool,
    source_code_retrieval_enabled: Optional[bool],
    retrieval_filter: AirflowFilter,
) -> "SerializedAirflowDefinitionsData":
    mapping_info = build_airlift_metadata_mapping_info(mapped_assets)
    fetched_airflow_data = fetch_all_airflow_data(
        airflow_instance,
        mapping_info,
        dag_selector_fn,
        automapping_enabled=automapping_enabled,
        retrieval_filter=retrieval_filter,
    )
    source_code_retrieval_enabled = infer_code_retrieval_enabled(
        source_code_retrieval_enabled, fetched_airflow_data
    )
    return SerializedAirflowDefinitionsData(
        instance_name=airflow_instance.name,
        key_scoped_task_handles=[
            KeyScopedTaskHandles(asset_key=k, mapped_tasks=v)
            for k, v in fetched_airflow_data.all_mapped_tasks.items()
        ],
        key_scoped_dag_handles=[
            KeyScopedDagHandles(asset_key=k, mapped_dags=v)
            for k, v in fetched_airflow_data.all_mapped_dags.items()
        ],
        dag_datas={
            dag_id: SerializedDagData(
                dag_id=dag_id,
                dag_info=dag_info,
                source_code=airflow_instance.get_dag_source_code(dag_info.metadata["file_token"])
                if source_code_retrieval_enabled
                else None,
                leaf_asset_keys=get_leaf_assets_for_dag(
                    asset_keys_in_dag=mapping_info.all_mapped_asset_keys_by_dag_id[dag_id],
                    downstreams_asset_dependency_graph=mapping_info.downstream_deps,
                ),
                task_infos=fetched_airflow_data.task_info_map[dag_id],
            )
            for dag_id, dag_info in fetched_airflow_data.dag_infos.items()
        },
        datasets=fetched_airflow_data.datasets,
    )



import logging
import os
import time
from abc import ABC, abstractmethod
from collections.abc import Iterable, Mapping, Sequence
from datetime import datetime
from typing import Any

import requests
from airflow.models import BaseOperator
from requests import Response

from dagster_airlift.constants import DAG_ID_TAG_KEY, DAG_RUN_ID_TAG_KEY, TASK_ID_TAG_KEY
from dagster_airlift.in_airflow.dagster_run_utils import DagsterRunResult
from dagster_airlift.in_airflow.gql_queries import (
    ASSET_NODES_QUERY,
    RUNS_QUERY,
    TRIGGER_ASSETS_MUTATION,
    VERIFICATION_QUERY,
)
from dagster_airlift.in_airflow.partition_utils import (
    PARTITION_NAME_TAG,
    PartitioningInformation,
    translate_logical_date_to_partition_key,
)

logger = logging.getLogger(__name__)

# A job in dagster is uniquely defined by (location_name, repository_name, job_name).
DagsterJobIdentifier = tuple[str, str, str]

# Airflow context type. Instead of using a strongly typed `Context` from Airflow 2.x,
# it is defined as a generic mapping for compatibility with Airflow 1.x
Context = Mapping[str, Any]

IMPLICIT_ASSET_JOB_PREFIX = "__ASSET_JOB"

DEFAULT_DAGSTER_RUN_STATUS_POLL_INTERVAL = 1


class BaseDagsterAssetsOperator(BaseOperator, ABC):
    """Interface for an operator which materializes dagster assets.

    This operator needs to implement the following methods:

        - get_dagster_session: Returns a requests session that can be used to make requests to the Dagster API.
            This is where any additional authentication can be added.
        - get_dagster_url: Returns the URL for the Dagster instance.
        - filter_asset_nodes: Filters asset nodes (which are returned from Dagster's graphql API) to only include those
            that should be triggered by the current task.

    Optionally, these methods can be overridden as well:

        - get_partition_key: Determines the partition key to use to trigger the dagster run. This method will only be
            called if the underlying asset is partitioned.
    """

    def __init__(
        self,
        dagster_run_status_poll_interval: int = DEFAULT_DAGSTER_RUN_STATUS_POLL_INTERVAL,
        **kwargs,
    ):
        super().__init__(**kwargs)
        self.dagster_run_status_poll_interval = dagster_run_status_poll_interval

    @abstractmethod
    def get_dagster_session(self, context: Context) -> requests.Session:
        """Returns a requests session that can be used to make requests to the Dagster API."""

    def _get_validated_session(self, context: Context) -> requests.Session:
        session = self.get_dagster_session(context)
        dagster_url = self.get_dagster_url(context)
        response = session.post(
            # Timeout in seconds
            f"{dagster_url}/graphql",
            json={"query": VERIFICATION_QUERY},
            timeout=3,
        )
        if response.status_code != 200:
            raise Exception(
                f"Failed to connect to Dagster at {dagster_url}. Response: {response.text}"
            )
        return session

    @abstractmethod
    def get_dagster_url(self, context: Context) -> str:
        """Returns the URL for the Dagster instance."""

    @abstractmethod
    def filter_asset_nodes(
        self, context: Context, asset_nodes: Sequence[Mapping[str, Any]]
    ) -> Iterable[Mapping[str, Any]]:
        """Filters the asset nodes to only include those that should be triggered by the current task."""

    def get_partition_key(
        self, context: Context, partitioning_info: PartitioningInformation
    ) -> str:
        """Overrideable method to determine the partition key to use to trigger the dagster run.

        This method will only be called if the underlying asset is partitioned.
        """
        if not partitioning_info:
            return None
        return translate_logical_date_to_partition_key(
            self.get_airflow_logical_date(context), partitioning_info
        )

    def get_valid_graphql_response(self, response: Response, key: str) -> Any:
        response_json = response.json()
        if not response_json.get("data"):
            raise Exception(f"Error in GraphQL request. No data key: {response_json}")

        if key not in response_json["data"]:
            raise Exception(f"Error in GraphQL request. No {key} key: {response_json}")

        return response_json["data"][key]

    def get_all_asset_nodes(
        self, session: requests.Session, dagster_url: str, context: Context
    ) -> Sequence[Mapping[str, Any]]:
        # create graphql client
        response = session.post(
            # Timeout in seconds
            f"{dagster_url}/graphql",
            json={"query": ASSET_NODES_QUERY},
            timeout=3,
        )
        return self.get_valid_graphql_response(response, "assetNodes")

    def launch_dagster_run(
        self,
        context: Context,
        session: requests.Session,
        dagster_url: str,
        execution_params: Mapping[str, Any],
    ) -> str:
        response = session.post(
            f"{dagster_url}/graphql",
            json={
                "query": TRIGGER_ASSETS_MUTATION,
                "variables": {"executionParams": execution_params},
            },
            # Timeout in seconds
            timeout=10,
        )
        launch_data = self.get_valid_graphql_response(response, "launchPipelineExecution")
        return launch_data["run"]["id"]

    def get_dagster_run_obj(
        self, session: requests.Session, dagster_url: str, run_id: str
    ) -> Mapping[str, Any]:
        response = session.post(
            f"{dagster_url}/graphql",
            json={"query": RUNS_QUERY, "variables": {"runId": run_id}},
            # Timeout in seconds
            timeout=3,
        )
        return self.get_valid_graphql_response(response, "runOrError")

    def get_attribute_from_airflow_context(self, context: Context, attribute: str) -> Any:
        if attribute not in context or context[attribute] is None:
            raise Exception(f"Attribute {attribute} not found in context.")
        return context[attribute]

    def get_airflow_dag_run_id(self, context: Context) -> str:
        return self.get_attribute_from_airflow_context(context, "dag_run").run_id

    def get_airflow_dag_id(self, context: Context) -> str:
        return self.get_attribute_from_airflow_context(context, "dag_run").dag_id

    def get_airflow_task_id(self, context: Context) -> str:
        return self.get_attribute_from_airflow_context(context, "task").task_id

    def get_airflow_logical_date(self, context: Context) -> datetime:
        return self.get_attribute_from_airflow_context(context, "logical_date")

    def default_dagster_run_tags(self, context: Context) -> dict[str, str]:
        return {
            DAG_ID_TAG_KEY: self.get_airflow_dag_id(context),
            DAG_RUN_ID_TAG_KEY: self.get_airflow_dag_run_id(context),
            TASK_ID_TAG_KEY: self.get_airflow_task_id(context),
        }

    def launch_runs_for_task(self, context: Context, dag_id: str, task_id: str) -> None:
        """Launches runs for the given task in Dagster."""
        session = self._get_validated_session(context)
        dagster_url = self.get_dagster_url(context)

        asset_nodes_data = self.get_all_asset_nodes(session, dagster_url, context)
        logger.info(f"Got response {asset_nodes_data}")
        filtered_asset_nodes = [
            asset_node
            for asset_node in self.filter_asset_nodes(context, asset_nodes_data)
            if _is_asset_node_executable(asset_node)
        ]
        if not filtered_asset_nodes:
            raise Exception(f"No asset nodes found to trigger for task {dag_id}.{task_id}")
        if (
            not len(
                {_get_implicit_job_identifier(asset_node) for asset_node in filtered_asset_nodes}
            )
            == 1
        ):
            raise Exception(
                "Could not find an implicit asset job that can trigger all assets in this task. "
                "This may mean that you need to upgrade your Dagster version (1.8 or later), which allows all assets to be materialized in a single run, "
                "or that the assets are not in the same code location. "
                "`dagster-airlift` expects that all assets mapped to a given task exist within the same code location, so that they can be executed by the same run."
            )

        job_identifier = _get_implicit_job_identifier(next(iter(filtered_asset_nodes)))
        asset_key_paths = [asset_node["assetKey"]["path"] for asset_node in filtered_asset_nodes]
        logger.info(f"Triggering run for {job_identifier} with assets {asset_key_paths}")
        tags = self.default_dagster_run_tags(context)
        partitioning_info = PartitioningInformation.from_asset_node_graphql(filtered_asset_nodes)
        if partitioning_info:
            tags[PARTITION_NAME_TAG] = self.get_partition_key(context, partitioning_info)
        logger.info(f"Using tags {tags}")
        run_id = self.launch_dagster_run(
            context,
            session,
            dagster_url,
            build_dagster_run_execution_params(
                tags,
                job_identifier,
                asset_key_paths=asset_key_paths,
            ),
        )
        logger.info("Waiting for dagster run completion...")
        self.wait_for_run_and_retries_to_complete(
            session=session, dagster_url=dagster_url, run_id=run_id
        )
        logger.info("All runs completed successfully.")
        return None

    def wait_for_run_to_complete(
        self, session: requests.Session, dagster_url: str, run_id: str
    ) -> DagsterRunResult:
        while True:
            response = self.get_dagster_run_obj(session, dagster_url, run_id)
            status = response["status"]
            if status in ["SUCCESS", "FAILURE", "CANCELED"]:
                break
            time.sleep(self.dagster_run_status_poll_interval)
        tags = {tag["key"]: tag["value"] for tag in response["tags"]}
        return DagsterRunResult(status=response["status"], tags=tags)

    def wait_for_run_and_retries_to_complete(
        self, session: requests.Session, dagster_url: str, run_id: str
    ) -> None:
        run_id_to_check = run_id
        while True:
            result = self.wait_for_run_to_complete(
                session=session, dagster_url=dagster_url, run_id=run_id_to_check
            )
            if result.succeeded:
                break
            logger.info(f"Run {run_id_to_check} completed with status '{result.status}'.")
            if result.run_will_automatically_retry and result.retried_run_id:
                logger.info(
                    f"Run {run_id_to_check} retried in run {result.retried_run_id}. Waiting for completion..."
                )
                run_id_to_check = result.retried_run_id
                continue
            elif result.run_will_automatically_retry:
                logger.info(
                    f"Run {run_id_to_check} failed, but is configured to automatically retry. Waiting for retried run to be created..."
                )
                continue
            else:
                raise Exception(f"Run {run_id_to_check} failed, and is not expected to retry.")
        return None

    def execute(self, context: Context) -> Any:
        # https://github.com/apache/airflow/discussions/24463
        os.environ["NO_PROXY"] = "*"
        dag_id = os.environ["AIRFLOW_CTX_DAG_ID"]
        task_id = os.environ["AIRFLOW_CTX_TASK_ID"]
        return self.launch_runs_for_task(context, dag_id, task_id)


def _get_implicit_job_identifier(asset_node: Mapping[str, Any]) -> DagsterJobIdentifier:
    """Extracts the implicit job identifier from an asset node.

    In dagster 1.8 and later, there is a single implicit asset job constructed across all assets.
    Using this job to execute assets allows us to minimize the number of runs we need to launch,
    and ensures that assets are executed in topological order.
    """
    # In dagster 1.8 and later, there is a single implicit asset job constructed across all assets. Using this job to execute the asset is preferred, because
    # it minimizes the number of runs we need to launch, and ensures that assets are executed
    implicit_asset_job = next(
        iter(
            [job for job in asset_node["jobs"] if job["name"].startswith(IMPLICIT_ASSET_JOB_PREFIX)]
        ),
        None,
    )
    job_to_use = implicit_asset_job or asset_node["jobs"][0]
    location_name = job_to_use["repository"]["location"]["name"]
    repository_name = job_to_use["repository"]["name"]
    job_name = job_to_use["name"]
    return (location_name, repository_name, job_name)


def build_dagster_run_execution_params(
    tags: Mapping[str, Any],
    job_identifier: DagsterJobIdentifier,
    asset_key_paths: Sequence[Sequence[str]],
) -> dict[str, Any]:
    location_name, repository_name, job_name = job_identifier
    return {
        "mode": "default",
        "executionMetadata": {
            "tags": [{"key": key, "value": value} for key, value in tags.items()]
        },
        "runConfigData": "{}",
        "selector": {
            "repositoryLocationName": location_name,
            "repositoryName": repository_name,
            "pipelineName": job_name,
            "assetSelection": [{"path": asset_key} for asset_key in asset_key_paths],
            "assetCheckSelection": [],
        },
    }


def _is_asset_node_executable(asset_node: Mapping[str, Any]) -> bool:
    return bool(asset_node["jobs"])


def _build_runs_filter_param(tags: Mapping[str, Any]) -> Mapping[str, Any]:
    return {"tags": [{"key": key, "value": value} for key, value in tags.items()]}



from collections import defaultdict
from collections.abc import Mapping, Sequence
from datetime import datetime, timedelta
from typing import Any, Optional

import requests

from dagster_airlift.core import AirflowInstance
from dagster_airlift.core.airflow_instance import DagInfo, TaskInfo
from dagster_airlift.core.basic_auth import AirflowAuthBackend
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance
from dagster_airlift.core.serialization.serialized_data import (
    Dataset,
    DatasetConsumingDag,
    DatasetProducingTask,
)


class DummyAuthBackend(AirflowAuthBackend):
    def get_session(self) -> requests.Session:
        raise NotImplementedError("This shouldn't be called from this mock context.")

    def get_webserver_url(self) -> str:
        return "http://dummy.domain"


DEFAULT_FAKE_INSTANCE_NAME = "test_instance"


class AirflowInstanceFake(AirflowInstance):
    """Loads a set of provided DagInfo, TaskInfo, and TaskInstance objects for testing."""

    def __init__(
        self,
        dag_infos: list[DagInfo],
        task_infos: list[TaskInfo],
        task_instances: list[TaskInstance],
        dag_runs: list[DagRun],
        datasets: list[Dataset] = [],
        variables: list[dict[str, Any]] = [],
        instance_name: Optional[str] = None,
        max_runs_per_batch: Optional[int] = None,
        logs: Optional[Mapping[str, Mapping[str, str]]] = None,
    ) -> None:
        self._dag_infos_by_dag_id = {dag_info.dag_id: dag_info for dag_info in dag_infos}
        self._task_infos_by_dag_and_task_id = {
            (task_info.dag_id, task_info.task_id): task_info for task_info in task_infos
        }
        self._task_instances_by_dag_and_task_id: dict[tuple[str, str], list[TaskInstance]] = (
            defaultdict(list)
        )
        self._logs_by_run_id_and_task_id: dict[tuple[str, str], str] = defaultdict(lambda: "")
        for run_id, task_log_map in (logs or {}).items():
            for task_id, log in task_log_map.items():
                self._logs_by_run_id_and_task_id[(run_id, task_id)] = log

        for task_instance in task_instances:
            self._task_instances_by_dag_and_task_id[
                (task_instance.dag_id, task_instance.task_id)
            ].append(task_instance)
        self._dag_runs_by_dag_id: dict[str, list[DagRun]] = defaultdict(list)
        for dag_run in dag_runs:
            self._dag_runs_by_dag_id[dag_run.dag_id].append(dag_run)
        self._dag_infos_by_file_token = {dag_info.file_token: dag_info for dag_info in dag_infos}
        self._variables = variables
        self._max_runs_per_batch = max_runs_per_batch
        self._datasets = datasets
        super().__init__(
            auth_backend=DummyAuthBackend(),
            name=DEFAULT_FAKE_INSTANCE_NAME if instance_name is None else instance_name,
        )

    def list_dags(self, retrieval_filter: Optional[AirflowFilter] = None) -> list[DagInfo]:
        retrieval_filter = retrieval_filter or AirflowFilter()
        # Very basic filtering for testing purposes
        dags_to_retrieve = list(self._dag_infos_by_dag_id.values())
        if retrieval_filter.dag_id_ilike:
            dags_to_retrieve = [
                dag_info
                for dag_info in dags_to_retrieve
                if retrieval_filter.dag_id_ilike in dag_info.dag_id
            ]
        if retrieval_filter.airflow_tags:
            dags_to_retrieve = [
                dag_info
                for dag_info in dags_to_retrieve
                if all(
                    tag in dag_info.metadata.get("tags", [])
                    for tag in retrieval_filter.airflow_tags
                )
            ]
        return dags_to_retrieve

    def list_variables(self) -> list[dict[str, Any]]:
        return self._variables

    def get_dag_runs(self, dag_id: str, start_date: datetime, end_date: datetime) -> list[DagRun]:
        if dag_id not in self._dag_runs_by_dag_id:
            raise ValueError(f"Dag run not found for dag_id {dag_id}")
        return [
            run
            for run in self._dag_runs_by_dag_id[dag_id]
            if start_date.timestamp() <= run.start_date.timestamp() <= end_date.timestamp()
            and start_date.timestamp() <= run.end_date.timestamp() <= end_date.timestamp()
        ]

    def _task_instance_flat_list(self) -> list[TaskInstance]:
        return [
            task_instance
            for task_instances in self._task_instances_by_dag_and_task_id.values()
            for task_instance in task_instances
        ]

    def get_dag_runs_batch(
        self,
        dag_ids: Sequence[str],
        end_date_gte: Optional[datetime] = None,
        end_date_lte: Optional[datetime] = None,
        start_date_gte: Optional[datetime] = None,
        start_date_lte: Optional[datetime] = None,
        offset: int = 0,
        states: Optional[Sequence[str]] = None,
    ) -> tuple[list[DagRun], int]:
        if end_date_gte and end_date_lte:
            runs = [
                (run.end_date, run)
                for runs in self._dag_runs_by_dag_id.values()
                for run in runs
                if (states is None or run.state in states)
                and end_date_gte.timestamp() <= run.end_date.timestamp() <= end_date_lte.timestamp()
                and run.dag_id in dag_ids
            ]
        elif start_date_gte and start_date_lte:
            runs = [
                (run.start_date, run)
                for runs in self._dag_runs_by_dag_id.values()
                for run in runs
                if start_date_gte.timestamp()
                <= run.start_date.timestamp()
                <= start_date_lte.timestamp()
                and run.dag_id in dag_ids
                and (states is None or run.state in states)
            ]
        else:
            raise ValueError(
                "Either end_date_gte and end_date_lte or start_date_gte and start_date_lte must be provided."
            )
        sorted_runs = [run for _, run in sorted(runs, key=lambda x: x[0])]
        end_idx = (
            offset + self._max_runs_per_batch if self._max_runs_per_batch else len(sorted_runs)
        )
        return (sorted_runs[offset:end_idx], len(sorted_runs))

    def get_task_instance_batch_time_range(
        self,
        dag_ids: Sequence[str],
        states: Sequence[str],
        end_date_gte: datetime,
        end_date_lte: datetime,
    ) -> list["TaskInstance"]:
        return [
            task_instance
            for task_instance in self._task_instance_flat_list()
            if end_date_gte.timestamp()
            <= task_instance.end_date.timestamp()
            <= end_date_lte.timestamp()
            and task_instance.dag_id in dag_ids
            and task_instance.state in states
        ]

    def get_task_instance_batch(
        self, dag_id: str, task_ids: Sequence[str], run_id: str, states: Sequence[str]
    ) -> list[TaskInstance]:
        task_instances = []
        for task_id in set(task_ids):
            if (dag_id, task_id) not in self._task_instances_by_dag_and_task_id:
                continue
            task_instances.extend(
                [
                    task_instance
                    for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
                    if task_instance.run_id == run_id and task_instance.state in states
                ]
            )
        return task_instances

    def get_task_instance(self, dag_id: str, task_id: str, run_id: str) -> TaskInstance:
        if (dag_id, task_id) not in self._task_instances_by_dag_and_task_id:
            raise ValueError(f"Task instance not found for dag_id {dag_id} and task_id {task_id}")
        if not any(
            task_instance.run_id == run_id
            for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
        ):
            raise ValueError(
                f"Task instance not found for dag_id {dag_id}, task_id {task_id}, and run_id {run_id}"
            )
        return next(
            iter(
                task_instance
                for task_instance in self._task_instances_by_dag_and_task_id[(dag_id, task_id)]
                if task_instance.run_id == run_id
            )
        )

    def get_task_info(self, dag_id, task_id) -> TaskInfo:  # pyright: ignore[reportIncompatibleMethodOverride]
        if (dag_id, task_id) not in self._task_infos_by_dag_and_task_id:
            raise ValueError(f"Task info not found for dag_id {dag_id} and task_id {task_id}")
        return self._task_infos_by_dag_and_task_id[(dag_id, task_id)]

    def get_task_infos(self, *, dag_id: str) -> list[TaskInfo]:
        if dag_id not in self._dag_infos_by_dag_id:
            raise ValueError(f"Dag info not found for dag_id {dag_id}")
        task_infos = []
        for dag_id, task_id in self._task_infos_by_dag_and_task_id:
            if dag_id == dag_id:
                task_infos.append(self._task_infos_by_dag_and_task_id[(dag_id, task_id)])
        return task_infos

    def get_dag_info(self, dag_id) -> DagInfo:
        if dag_id not in self._dag_infos_by_dag_id:
            raise ValueError(f"Dag info not found for dag_id {dag_id}")
        return self._dag_infos_by_dag_id[dag_id]

    def get_dag_source_code(self, file_token: str) -> str:
        if file_token not in self._dag_infos_by_file_token:
            raise ValueError(f"Dag info not found for file_token {file_token}")
        return "indicates found source code"

    def get_all_datasets(
        self,
        *,
        batch_size=100,
        retrieval_filter: Optional[AirflowFilter] = None,
        dag_ids: Optional[Sequence[str]] = None,
    ) -> Sequence[Dataset]:
        return_datasets = []
        retrieval_filter = retrieval_filter or AirflowFilter()
        for dataset in self._datasets:
            if (
                retrieval_filter.dataset_uri_ilike
                and retrieval_filter.dataset_uri_ilike not in dataset.uri
            ):
                continue
            if dag_ids and not any(t.dag_id in dag_ids for t in dataset.producing_tasks):
                continue
            return_datasets.append(dataset)
        return return_datasets

    def get_task_instance_logs(
        self, dag_id: str, task_id: str, run_id: str, try_number: int
    ) -> str:
        return self._logs_by_run_id_and_task_id[(run_id, task_id)]


def make_dag_info(
    instance_name: str, dag_id: str, file_token: Optional[str], dag_props: Mapping[str, Any]
) -> DagInfo:
    return DagInfo(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        metadata={"file_token": file_token if file_token else "dummy_file_token", **dag_props},
    )


def make_task_info(
    dag_id: str, task_id: str, downstream_task_ids: Optional[list[str]] = None
) -> TaskInfo:
    return TaskInfo(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        task_id=task_id,
        metadata={"downstream_task_ids": downstream_task_ids or []},
    )


def make_task_instance(
    dag_id: str,
    task_id: str,
    run_id: str,
    start_date: datetime,
    end_date: datetime,
    logical_date: Optional[datetime] = None,
    try_number: int = 1,
) -> TaskInstance:
    return TaskInstance(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        task_id=task_id,
        run_id=run_id,
        metadata={
            "state": "success",
            "start_date": start_date.isoformat(),
            "end_date": end_date.isoformat(),
            "logical_date": logical_date.isoformat() if logical_date else start_date.isoformat(),
            "try_number": try_number,
        },
    )


def make_dag_run(
    dag_id: str,
    run_id: str,
    start_date: datetime,
    end_date: Optional[datetime],
    logical_date: Optional[datetime] = None,
    state: Optional[str] = None,
) -> DagRun:
    metadata = {
        "state": state or "success",
        "start_date": start_date.isoformat(),
        "logical_date": logical_date.isoformat() if logical_date else start_date.isoformat(),
        "run_type": "manual",
        "note": "dummy note",
        "conf": {},
    }
    if end_date:
        metadata["end_date"] = end_date.isoformat()
    return DagRun(
        webserver_url="http://dummy.domain",
        dag_id=dag_id,
        run_id=run_id,
        metadata=metadata,
    )


def make_dataset(
    producing_tasks: Sequence[Mapping[str, str]],
    consuming_dags: Sequence[str],
    uri: str,
    extra: Mapping[str, Any],
) -> Dataset:
    return Dataset(
        id=1,
        uri=uri,
        extra=extra,
        created_at=datetime.now().isoformat(),
        updated_at=datetime.now().isoformat(),
        producing_tasks=[
            DatasetProducingTask(
                dag_id=task["dag_id"],
                task_id=task["task_id"],
                created_at=datetime.now().isoformat(),
                updated_at=datetime.now().isoformat(),
            )
            for task in producing_tasks
        ],
        consuming_dags=[
            DatasetConsumingDag(
                dag_id=dag_id,
                created_at=datetime.now().isoformat(),
                updated_at=datetime.now().isoformat(),
            )
            for dag_id in consuming_dags
        ],
    )


def make_instance(
    dag_and_task_structure: dict[str, list[str]],
    dataset_construction_info: Sequence[Mapping[str, Any]] = [],
    dag_runs: list[DagRun] = [],
    task_deps: dict[str, list[str]] = {},
    instance_name: Optional[str] = None,
    max_runs_per_batch: Optional[int] = None,
    dag_props: dict[str, Any] = {},
    task_instances: Optional[list[TaskInstance]] = None,
    logs: Optional[Mapping[str, Mapping[str, str]]] = None,
) -> AirflowInstanceFake:
    """Constructs DagInfo, TaskInfo, and TaskInstance objects from provided data.

    Args:
        dag_and_task_structure: A dictionary mapping dag_id to a list of task_ids.
        dag_runs: A list of DagRun objects to include in the instance. A TaskInstance object will be
            created for each task_id in the dag, for each DagRun in dag_runs pertaining to a particular dag.
    """
    dag_infos = []
    task_infos = []
    for dag_id, task_ids in dag_and_task_structure.items():
        dag_info = make_dag_info(
            instance_name=instance_name or DEFAULT_FAKE_INSTANCE_NAME,
            dag_id=dag_id,
            file_token=dag_id,
            dag_props=dag_props.get(dag_id, {}),
        )
        dag_infos.append(dag_info)
        task_infos.extend(
            [
                make_task_info(
                    dag_id=dag_id, task_id=task_id, downstream_task_ids=task_deps.get(task_id, [])
                )
                for task_id in task_ids
            ]
        )
    if not task_instances:
        task_instances = []
        for dag_run in dag_runs:
            task_instances.extend(
                [
                    make_task_instance(
                        dag_id=dag_run.dag_id,
                        task_id=task_id,
                        run_id=dag_run.run_id,
                        start_date=dag_run.start_date,
                        end_date=dag_run.end_date
                        - timedelta(
                            seconds=1
                        ),  # Ensure that the task ends before the full "dag" completes.
                        logical_date=dag_run.logical_date,
                    )
                    for task_id in dag_and_task_structure[dag_run.dag_id]
                ]
            )
    datasets = []
    for dataset_info in dataset_construction_info:
        datasets.append(
            make_dataset(
                producing_tasks=dataset_info["producing_tasks"],
                consuming_dags=dataset_info["consuming_dags"],
                uri=dataset_info["uri"],
                extra=dataset_info.get("extra", {}),
            )
        )
    return AirflowInstanceFake(
        dag_infos=dag_infos,
        task_infos=task_infos,
        task_instances=task_instances,
        dag_runs=dag_runs,
        instance_name=instance_name,
        max_runs_per_batch=max_runs_per_batch,
        datasets=datasets,
        logs=logs,
    )



import logging
from collections.abc import Callable
from typing import Any, Optional

from airflow import DAG
from airflow.models import BaseOperator

from dagster_airlift.in_airflow.dag_proxy_operator import (
    BaseProxyDAGToDagsterOperator,
    DefaultProxyDAGToDagsterOperator,
)
from dagster_airlift.in_airflow.proxied_state import AirflowProxiedState
from dagster_airlift.in_airflow.task_proxy_operator import (
    BaseProxyTaskToDagsterOperator,
    DefaultProxyTaskToDagsterOperator,
)


def proxying_to_dagster(
    *,
    global_vars: dict[str, Any],
    proxied_state: AirflowProxiedState,
    logger: Optional[logging.Logger] = None,
    build_from_task_fn: Callable[
        [BaseOperator], BaseProxyTaskToDagsterOperator
    ] = DefaultProxyTaskToDagsterOperator.build_from_task,
    build_from_dag_fn: Callable[
        [DAG], BaseProxyDAGToDagsterOperator
    ] = DefaultProxyDAGToDagsterOperator.build_from_dag,
) -> None:
    """Proxies tasks and dags to Dagster based on provided proxied state.
    Expects a dictionary of in-scope global variables to be provided (typically retrieved with `globals()`), and a proxied state dictionary
    (typically retrieved with :py:func:`load_proxied_state_from_yaml`) for dags in that global state. This function will modify in-place the
    dictionary of global variables to replace proxied tasks with appropriate Dagster operators.

    In the case of task-level proxying, the proxied tasks will be replaced with new operators that are constructed by the provided `build_from_task_fn`.
    A default implementation of this function is provided in `DefaultProxyTaskToDagsterOperator`.
    In the case of dag-level proxying, the entire dag structure will be replaced with a single task that is constructed by the provided `build_from_dag_fn`.
    A default implementation of this function is provided in `DefaultProxyDAGToDagsterOperator`.


    Args:
        global_vars (Dict[str, Any]): The global variables in the current context. In most cases, retrieved with `globals()` (no import required).
            This is equivalent to what airflow already does to introspect the dags which exist in a given module context:
            https://airflow.apache.org/docs/apache-airflow/stable/core-concepts/dags.html#loading-dags
        proxied_state (AirflowMigrationState): The proxied state for the dags.
        logger (Optional[logging.Logger]): The logger to use. Defaults to logging.getLogger("dagster_airlift").


    Examples:
        Typical usage of this function is to be called at the end of a dag file, retrieving proxied_state from an accompanying `proxied_state` path.

        .. code-block:: python

            from pathlib import Path

            from airflow import DAG
            from airflow.operators.python import PythonOperator
            from dagster._time import get_current_datetime_midnight
            from dagster_airlift.in_airflow import proxying_to_dagster
            from dagster_airlift.in_airflow.proxied_state import load_proxied_state_from_yaml


            with DAG(
                dag_id="daily_interval_dag",
                ...,
            ) as minute_dag:
                PythonOperator(task_id="my_task", python_callable=...)

            # At the end of the dag file, so we can ensure dags are loaded into globals.
            proxying_to_dagster(
                proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
                global_vars=globals(),
            )

        You can also provide custom implementations of the `build_from_task_fn` function to customize the behavior of task-level proxying.

        .. code-block:: python

            from dagster_airlift.in_airflow import proxying_to_dagster, BaseProxyTaskToDagsterOperator
            from airflow.models.operator import BaseOperator

            ... # Dag code here

            class CustomAuthTaskProxyOperator(BaseProxyTaskToDagsterOperator):
                def get_dagster_session(self, context: Context) -> requests.Session:
                    # Add custom headers to the session
                    return requests.Session(headers={"Authorization": "Bearer my_token"})

                def get_dagster_url(self, context: Context) -> str:
                    # Use a custom environment variable for the dagster url
                    return os.environ["CUSTOM_DAGSTER_URL"]

                @classmethod
                def build_from_task(cls, task: BaseOperator) -> "CustomAuthTaskProxyOperator":
                    # Custom logic to build the operator from the task (task_id should remain the same)
                    if task.task_id == "my_task_needs_more_retries":
                        return CustomAuthTaskProxyOperator(task_id=task_id, retries=3)
                    else:
                        return CustomAuthTaskProxyOperator(task_id=task_id)

            proxying_to_dagster(
                proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
                global_vars=globals(),
                build_from_task_fn=CustomAuthTaskProxyOperator.build_from_task,
            )

        You can do the same for dag-level proxying by providing a custom implementation of the `build_from_dag_fn` function.

        .. code-block:: python

            from dagster_airlift.in_airflow import proxying_to_dagster, BaseProxyDAGToDagsterOperator
            from airflow.models.dag import DAG

            ... # Dag code here

            class CustomAuthDAGProxyOperator(BaseProxyDAGToDagsterOperator):
                def get_dagster_session(self, context: Context) -> requests.Session:
                    # Add custom headers to the session
                    return requests.Session(headers={"Authorization": "Bearer my_token"})

                def get_dagster_url(self, context: Context) -> str:
                    # Use a custom environment variable for the dagster url
                    return os.environ["CUSTOM_DAGSTER_URL"]

                @classmethod
                def build_from_dag(cls, dag: DAG) -> "CustomAuthDAGProxyOperator":
                    # Custom logic to build the operator from the dag (DAG id should remain the same)
                    if dag.dag_id == "my_dag_needs_more_retries":
                        return CustomAuthDAGProxyOperator(task_id="custom override", retries=3, dag=dag)
                    else:
                        return CustomAuthDAGProxyOperator(task_id="basic_override", dag=dag)

            proxying_to_dagster(
                proxied_state=load_proxied_state_from_yaml(Path(__file__).parent / "proxied_state"),
                global_vars=globals(),
                build_from_dag_fn=CustomAuthDAGProxyOperator.build_from_dag,
            )

    """
    caller_module = global_vars.get("__module__")
    suffix = f" in module `{caller_module}`" if caller_module else ""
    if not logger:
        logger = logging.getLogger("dagster_airlift")
    logger.debug(f"Searching for dags proxied to dagster{suffix}...")
    task_level_proxying_dags: list[DAG] = []
    dag_level_proxying_dags: list[DAG] = []
    all_dag_ids: set[str] = set()
    # Do a pass to collect dags and ensure that proxied information is set correctly.
    for obj in global_vars.values():
        if not isinstance(obj, DAG):
            continue
        dag: DAG = obj
        all_dag_ids.add(dag.dag_id)
        if not proxied_state.dag_has_proxied_state(dag.dag_id):
            logger.debug(f"Dag with id `{dag.dag_id}` has no proxied state. Skipping...")
            continue
        logger.debug(f"Dag with id `{dag.dag_id}` has proxied state.")
        proxied_state_for_dag = proxied_state.dags[dag.dag_id]
        if proxied_state_for_dag.proxied is not None:
            if proxied_state_for_dag.proxied is False:
                logger.debug(f"Dag with id `{dag.dag_id}` is not proxied. Skipping...")
                continue
            dag_level_proxying_dags.append(dag)
        else:
            for task_id in proxied_state_for_dag.tasks.keys():
                if task_id not in dag.task_dict:
                    raise Exception(
                        f"Task with id `{task_id}` not found in dag `{dag.dag_id}`. Found tasks: {list(dag.task_dict.keys())}"
                    )
                if not isinstance(dag.task_dict[task_id], BaseOperator):
                    raise Exception(
                        f"Task with id `{task_id}` in dag `{dag.dag_id}` is not an instance of BaseOperator. This likely means a MappedOperator was attempted, which is not yet supported by airlift."
                    )
            task_level_proxying_dags.append(dag)

    if len(all_dag_ids) == 0:
        raise Exception(
            "No dags found in globals dictionary. Ensure that your dags are available from global context, and that the call to `proxying_to_dagster` is the last line in your dag file."
        )

    for dag in dag_level_proxying_dags:
        logger.debug(f"Tagging dag {dag.dag_id} as proxied.")
        dag.tags = [*dag.tags, "Dag overriden to Dagster"]
        dag.task_dict = {}
        dag.task_group.children = {}
        override_task = build_from_dag_fn(dag)
        dag.task_dict[override_task.task_id] = override_task
    for dag in task_level_proxying_dags:
        logger.debug(f"Tagging dag {dag.dag_id} as proxied.")
        proxied_state_for_dag = proxied_state.dags[dag.dag_id]
        num_proxied_tasks = len(
            [
                task_id
                for task_id, task_state in proxied_state_for_dag.tasks.items()
                if task_state.proxied
            ]
        )
        task_possessive = "Task" if num_proxied_tasks == 1 else "Tasks"
        dag.tags = [
            *dag.tags,
            f"{num_proxied_tasks} {task_possessive} Marked as Proxied to Dagster",
        ]
        proxied_tasks = set()
        for task_id, task_state in proxied_state_for_dag.tasks.items():
            if not task_state.proxied:
                logger.debug(
                    f"Task {task_id} in dag {dag.dag_id} has `proxied` set to False. Skipping..."
                )
                continue

            # At this point, we should be assured that the task exists within the task_dict of the dag, and is a BaseOperator.
            original_op: BaseOperator = dag.task_dict[task_id]  # type: ignore  # we already confirmed this is BaseOperator
            del dag.task_dict[task_id]
            if original_op.task_group is not None:
                del original_op.task_group.children[task_id]
            logger.debug(
                f"Creating new operator for task {original_op.task_id} in dag {original_op.dag_id}"
            )
            new_op = build_from_task_fn(original_op)
            original_op.dag.task_dict[original_op.task_id] = new_op

            new_op.upstream_task_ids = original_op.upstream_task_ids
            new_op.downstream_task_ids = original_op.downstream_task_ids
            new_op.dag = original_op.dag
            original_op.dag = None
            proxied_tasks.add(task_id)
        logger.debug(f"Proxied tasks {proxied_tasks} in dag {dag.dag_id}.")
    logging.debug(f"Proxied {len(task_level_proxying_dags)}.")
    logging.debug(f"Completed switching proxied tasks to dagster{suffix}.")



from collections.abc import Mapping, Sequence
from typing import AbstractSet, Any, Callable  # noqa: UP035

from dagster import AssetKey, AssetSpec, JsonMetadataValue, UrlMetadataValue
from dagster._core.definitions.assets.definition.assets_definition import AssetsDefinition
from dagster._core.definitions.external_asset import external_asset_from_spec
from dagster._core.storage.tags import KIND_PREFIX

from dagster_airlift.constants import AUTOMAPPED_TASK_METADATA_KEY
from dagster_airlift.core.dag_asset import (
    dag_asset_metadata,
    dag_description,
    peered_dag_asset_metadata,
)
from dagster_airlift.core.serialization.serialized_data import (
    DagHandle,
    SerializedAirflowDefinitionsData,
    SerializedDagData,
    TaskHandle,
    TaskInfo,
)
from dagster_airlift.core.utils import (
    airflow_kind_dict,
    airlift_mapped_kind_dict,
    convert_to_valid_dagster_name,
    metadata_for_task_mapping,
)


def metadata_for_mapped_tasks(
    tasks: AbstractSet[TaskHandle], serialized_data: SerializedAirflowDefinitionsData
) -> Mapping[str, Any]:
    mapped_task = next(iter(tasks))
    task_info = serialized_data.dag_datas[mapped_task.dag_id].task_infos[mapped_task.task_id]
    task_level_metadata = {
        "Task Info (raw)": JsonMetadataValue(task_info.metadata),
        "Dag ID": task_info.dag_id,
        "Link to DAG": UrlMetadataValue(task_info.dag_url),
    }
    task_level_metadata["Task ID"] = task_info.task_id
    return task_level_metadata


def enrich_spec_with_airflow_task_metadata(
    spec: AssetSpec,
    tasks: AbstractSet[TaskHandle],
    serialized_data: SerializedAirflowDefinitionsData,
) -> AssetSpec:
    return spec.merge_attributes(
        metadata=metadata_for_mapped_tasks(tasks, serialized_data),
        tags=airlift_mapped_kind_dict(),
    )


def metadata_for_mapped_dags(
    dags: AbstractSet[DagHandle], serialized_data: SerializedAirflowDefinitionsData
) -> Mapping[str, Any]:
    mapped_dag = next(iter(dags))
    dag_info = serialized_data.dag_datas[mapped_dag.dag_id].dag_info
    return dag_asset_metadata(dag_info)


def enrich_spec_with_airflow_dag_metadata(
    spec: AssetSpec,
    dags: AbstractSet[DagHandle],
    serialized_data: SerializedAirflowDefinitionsData,
) -> AssetSpec:
    return spec.merge_attributes(
        metadata=metadata_for_mapped_dags(dags, serialized_data),
        tags=airlift_mapped_kind_dict(),
    )


def make_dag_external_asset(instance_name: str, dag_data: SerializedDagData) -> AssetsDefinition:
    return external_asset_from_spec(
        AssetSpec(
            key=make_default_dag_asset_key(instance_name, dag_data.dag_id),
            description=dag_description(dag_data.dag_info),
            metadata=peered_dag_asset_metadata(dag_data.dag_info, dag_data.source_code),
            tags={**airflow_kind_dict(), f"{KIND_PREFIX}dag": ""},
            deps=dag_data.leaf_asset_keys,
        )
    )


def get_airflow_data_to_spec_mapper(
    serialized_data: SerializedAirflowDefinitionsData,
) -> Callable[[AssetSpec], AssetSpec]:
    """Creates a mapping function s.t. if there is airflow data applicable to the asset key, transform the spec and apply the data."""

    def _fn(spec: AssetSpec) -> AssetSpec:
        if spec.key in serialized_data.all_mapped_tasks:
            return enrich_spec_with_airflow_task_metadata(
                spec, serialized_data.all_mapped_tasks[spec.key], serialized_data
            )
        elif spec.key in serialized_data.all_mapped_dags:
            return enrich_spec_with_airflow_dag_metadata(
                spec, serialized_data.all_mapped_dags[spec.key], serialized_data
            )
        return spec

    return _fn


def construct_dag_assets_defs(
    serialized_data: SerializedAirflowDefinitionsData,
) -> Sequence[AssetsDefinition]:
    return [
        make_dag_external_asset(serialized_data.instance_name, dag_data)
        for dag_data in serialized_data.dag_datas.values()
    ]


def key_for_automapped_task_asset(instance_name, dag_id, task_id) -> AssetKey:
    return AssetKey([instance_name, "dag", dag_id, "task", task_id])


def description_for_automapped_task_asset(task_info: TaskInfo) -> str:
    return f'Automapped task in dag "{task_info.dag_id}" with task_id "{task_info.task_id}"'


def tags_for_automapped_task_asset() -> Mapping[str, str]:
    return {f"{KIND_PREFIX}airflow": "", f"{KIND_PREFIX}task": ""}


def metadata_for_automapped_task_asset(task_info: TaskInfo) -> Mapping[str, Any]:
    return {
        **{
            "Task Info (raw)": JsonMetadataValue(task_info.metadata),
            "Dag ID": task_info.dag_id,
            "Task ID": task_info.task_id,
            "Link to DAG": UrlMetadataValue(task_info.dag_url),
            AUTOMAPPED_TASK_METADATA_KEY: True,
        },
        **metadata_for_task_mapping(task_id=task_info.task_id, dag_id=task_info.dag_id),
    }


def construct_automapped_dag_assets_defs(
    serialized_data: SerializedAirflowDefinitionsData,
) -> Sequence[AssetsDefinition]:
    dag_specs = []
    task_specs = []
    for dag_data in serialized_data.dag_datas.values():
        leaf_tasks = set()
        upstream_deps: dict[str, set[str]] = {task_id: set() for task_id in dag_data.task_infos}
        for task_id, task_info in dag_data.task_infos.items():
            if not task_info.downstream_task_ids:
                leaf_tasks.add(task_id)
            for downstream_id in task_info.downstream_task_ids:
                upstream_deps[downstream_id].add(task_id)

        task_specs.extend(
            AssetSpec(
                key=key_for_automapped_task_asset(
                    serialized_data.instance_name, dag_data.dag_id, task_id
                ),
                deps=[
                    key_for_automapped_task_asset(
                        serialized_data.instance_name, dag_data.dag_id, upstream_task_id
                    )
                    for upstream_task_id in upstream_task_ids
                ],
                description=description_for_automapped_task_asset(dag_data.task_infos[task_id]),
                tags=tags_for_automapped_task_asset(),
                metadata=metadata_for_automapped_task_asset(dag_data.task_infos[task_id]),
            )
            for task_id, upstream_task_ids in upstream_deps.items()
        )

        dag_specs.append(
            AssetSpec(
                key=make_default_dag_asset_key(serialized_data.instance_name, dag_data.dag_id),
                description=dag_description(dag_data.dag_info),
                metadata=peered_dag_asset_metadata(dag_data.dag_info, dag_data.source_code),
                tags=airflow_kind_dict(),
                deps=set(
                    [
                        key_for_automapped_task_asset(
                            serialized_data.instance_name, dag_data.dag_id, task_id
                        )
                        for task_id in leaf_tasks
                    ]
                ).union(dag_data.leaf_asset_keys),
            )
        )

    return task_specs + dag_specs


def make_default_dag_asset_key(instance_name: str, dag_id: str) -> AssetKey:
    """Conventional asset key representing a successful run of an airfow dag."""
    return AssetKey([instance_name, "dag", convert_to_valid_dagster_name(dag_id)])



import json
from collections.abc import Iterable, Sequence
from typing import Optional, TypeVar, Union, cast

import dagster._check as check
from dagster._core.definitions.metadata.external_metadata import (
    EXTERNAL_METADATA_TYPE_INFER,
    EXTERNAL_METADATA_TYPES,
    EXTERNAL_METADATA_VALUE_KEYS,
    ExternalMetadataValue,
    metadata_map_from_external,
)
from dagster._core.definitions.metadata.metadata_value import MetadataValue
from dagster._core.execution.context.op_execution_context import OpExecutionContext
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._core.storage.tags import EXTERNALLY_MANAGED_ASSETS_TAG

from dagster_airlift.constants import DAG_ID_TAG_KEY, DAG_RUN_ID_TAG_KEY, OBSERVATION_RUN_TAG_KEY
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance
from dagster_airlift.core.serialization.serialized_data import DagHandle, TaskHandle


def structured_log(context: OpExecutionContext, message: str) -> None:
    context.log.info(f"[Airflow Monitoring Job]: {message}")


def get_dagster_run_for_airflow_repr(
    context: OpExecutionContext, airflow_repr: Union[DagRun, TaskInstance]
) -> Optional[DagsterRun]:
    return next(
        iter(
            context.instance.get_runs(
                filters=RunsFilter(
                    tags={
                        DAG_RUN_ID_TAG_KEY: airflow_repr.run_id,
                        DAG_ID_TAG_KEY: airflow_repr.dag_id,
                        OBSERVATION_RUN_TAG_KEY: "true",
                    }
                ),
            )
        ),
        None,
    )


def get_externally_managed_runs_from_handle(
    context: OpExecutionContext, handle: Union[DagHandle, TaskHandle], run_id: str
) -> Sequence[DagsterRun]:
    return context.instance.get_runs(
        filters=RunsFilter(
            tags={
                **handle.identifying_tags,
                DAG_RUN_ID_TAG_KEY: run_id,
                EXTERNALLY_MANAGED_ASSETS_TAG: "true",
            },
        )
    )


_T = TypeVar("_T")


def _assert_param_value(value: _T, expected_values: Iterable[_T]) -> _T:
    if value not in expected_values:
        raise Exception(
            f"Invalid value when translating metadata from logs. Expected one of"
            f" `{expected_values}`, got `{value}`."
        )
    return value


def extract_metadata_from_logs(context: OpExecutionContext, logs: str) -> dict[str, MetadataValue]:
    metadata = {}
    import re

    matches = re.findall(r"DAGSTER_START(.*?)DAGSTER_END", logs, re.DOTALL)
    for match in matches:
        try:
            raw_external_metadata_map = json.loads(match)
        except json.JSONDecodeError as e:
            raise Exception(f"Invalid JSON found in logs for match: {match}. JSON error: {e}")
        check.mapping_param(raw_external_metadata_map, "raw_external_metadata_map")
        new_external_metadata_map = {}
        for key, value in raw_external_metadata_map.items():
            if not isinstance(key, str):
                raise Exception(
                    f"Invalid type when translating metadata from logs. Expected a dict with string"
                    f" keys, got a key `{key}` of type `{type(key)}`."
                )
            elif isinstance(value, dict):
                if not {*value.keys()} == EXTERNAL_METADATA_VALUE_KEYS:
                    raise Exception(
                        f"Invalid type when translating metadata from logs. Expected a dict with"
                        " string keys and values that are either raw metadata values or dictionaries"
                        f" with schema `{{raw_value: ..., type: ...}}`. Got a value `{value}`."
                    )
                _assert_param_value(value["type"], EXTERNAL_METADATA_TYPES)
                new_external_metadata_map[key] = cast("ExternalMetadataValue", value)
            else:
                new_external_metadata_map[key] = {
                    "raw_value": value,
                    "type": EXTERNAL_METADATA_TYPE_INFER,
                }

        metadata_map = metadata_map_from_external(new_external_metadata_map)
        metadata.update(metadata_map)

    return metadata



import datetime
from typing import TYPE_CHECKING, Any

from dagster import _check as check
from dagster._record import record

if TYPE_CHECKING:
    from dagster_airlift.core.serialization.serialized_data import DagHandle, TaskHandle


@record
class DagRun:
    webserver_url: str
    dag_id: str
    run_id: str
    metadata: dict[str, Any]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def dag_handle(self) -> "DagHandle":
        from dagster_airlift.core.serialization.serialized_data import DagHandle

        return DagHandle(dag_id=self.dag_id)

    @property
    def url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&tab=details"

    @property
    def success(self) -> bool:
        return self.metadata["state"] == "success"

    @property
    def finished(self) -> bool:
        from dagster_airlift.core.airflow_instance import TERMINAL_STATES

        return self.state in TERMINAL_STATES

    @property
    def state(self) -> str:
        return self.metadata["state"]

    @property
    def run_type(self) -> str:
        return self.metadata["run_type"]

    @property
    def config(self) -> dict[str, Any]:
        return self.metadata["conf"]

    @property
    def logical_date(self) -> datetime.datetime:
        """Returns the airflow-coined "logical date" from the dag run metadata.
        The logical date refers to the starting time of the "data interval" that the dag run is processing.
        In airflow < 2.2, this was set as the execution_date parameter in the dag run metadata.
        """
        # In airflow < 2.2, execution_date is set instead of logical_date.
        logical_date_str = check.not_none(
            self.metadata.get("logical_date") or self.metadata.get("execution_date"),
            "Expected one of execution_date or logical_date to be returned from the airflow rest API when querying for dag information.",
        )

        return datetime.datetime.fromisoformat(logical_date_str)

    @property
    def start_date(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.metadata["start_date"])

    @property
    def end_date(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.metadata["end_date"])


@record
class TaskInstance:
    webserver_url: str
    dag_id: str
    task_id: str
    run_id: str
    metadata: dict[str, Any]

    @property
    def state(self) -> str:
        return self.metadata["state"]

    @property
    def note(self) -> str:
        return self.metadata.get("note") or ""

    @property
    def details_url(self) -> str:
        return f"{self.webserver_url}/dags/{self.dag_id}/grid?dag_run_id={self.run_id}&task_id={self.task_id}"

    @property
    def log_url(self) -> str:
        return f"{self.details_url}&tab=logs"

    @property
    def logical_date(self) -> datetime.datetime:
        """Returns the airflow-coined "logical date" from the task instance metadata.
        The logical date refers to the starting time of the "data interval" that the overall dag run is processing.
        In airflow < 2.2, this was set as the execution_date parameter in the task instance metadata.
        """
        # In airflow < 2.2, execution_date is set instead of logical_date.
        logical_date_str = check.not_none(
            self.metadata.get("logical_date") or self.metadata.get("execution_date"),
            "Expected one of execution_date or logical_date to be returned from the airflow rest API when querying for task information.",
        )

        return datetime.datetime.fromisoformat(logical_date_str)

    @property
    def start_date(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.metadata["start_date"])

    @property
    def end_date(self) -> datetime.datetime:
        return datetime.datetime.fromisoformat(self.metadata["end_date"])

    @property
    def task_handle(self) -> "TaskHandle":
        from dagster_airlift.core.serialization.serialized_data import TaskHandle

        return TaskHandle(dag_id=self.dag_id, task_id=self.task_id)

    @property
    def dag_handle(self) -> "DagHandle":
        from dagster_airlift.core.serialization.serialized_data import DagHandle

        return DagHandle(dag_id=self.dag_id)

    @property
    def try_number(self) -> int:
        return self.metadata["try_number"]



import requests
from dagster._annotations import beta

from dagster_airlift.core.airflow_instance import AirflowAuthBackend


@beta
class AirflowBasicAuthBackend(AirflowAuthBackend):
    """A :py:class:`dagster_airlift.core.AirflowAuthBackend` that authenticates using basic auth.

    Args:
        webserver_url (str): The URL of the webserver.
        username (str): The username to authenticate with.
        password (str): The password to authenticate with.

    Examples:
        Creating a :py:class:`AirflowInstance` using this backend.

        .. code-block:: python

            from dagster_airlift.core import AirflowInstance, AirflowBasicAuthBackend

            af_instance = AirflowInstance(
                name="my-instance",
                auth_backend=AirflowBasicAuthBackend(
                    webserver_url="https://my-webserver-hostname",
                    username="my-username",
                    password="my-password"
                )
            )

    """

    def __init__(self, webserver_url: str, username: str, password: str):
        self._webserver_url = webserver_url
        self.username = username
        self.password = password

    def get_session(self) -> requests.Session:
        session = requests.Session()
        session.auth = (self.username, self.password)
        return session

    def get_webserver_url(self) -> str:
        return self._webserver_url



from collections.abc import Mapping, Sequence
from datetime import (
    datetime,
    timezone as tz,
)
from enum import Enum
from typing import Any, NamedTuple, Optional

PARTITION_NAME_TAG = "dagster/partition"


class PartitionDefinitionType(Enum):
    TIME_WINDOW = "TIME_WINDOW"
    STATIC = "STATIC"
    MULTIPARTITIONED = "MULTIPARTITIONED"
    DYNAMIC = "DYNAMIC"


class TimeWindowPartitioningInformation(NamedTuple):
    fmt: str


class PartitioningInformation(NamedTuple):
    partitioning_type: PartitionDefinitionType
    partition_keys: Sequence[str]
    # Eventually we can add more of these for different partitioning types
    additional_info: Optional[TimeWindowPartitioningInformation]

    @staticmethod
    def from_asset_node_graphql(
        asset_nodes: Sequence[Mapping[str, Any]],
    ) -> Optional["PartitioningInformation"]:
        assets_partitioned = [_asset_is_partitioned(asset_node) for asset_node in asset_nodes]
        if any(assets_partitioned) and not all(assets_partitioned):
            raise Exception(
                "Found some unpartitioned assets and some partitioned assets in the same task. "
                "For a given task, all assets must have the same partitions definition. "
            )
        partition_keys_per_asset = [
            set(asset_node["partitionKeys"])
            for asset_node in asset_nodes
            if asset_node["isPartitioned"]
        ]
        if not all_sets_equal(partition_keys_per_asset):
            raise Exception(
                "Found differing partition keys across assets in this task. "
                "For a given task, all assets must have the same partitions definition. "
            )
        # Now we can proceed with the assumption that all assets are partitioned and have the same partition keys.
        # This, we only look at the first asset node.
        asset_node = next(iter(asset_nodes))
        if not asset_node["isPartitioned"]:
            return None
        partitioning_type = PartitionDefinitionType(asset_node["partitionDefinition"]["type"])
        return PartitioningInformation(
            partitioning_type=partitioning_type,
            partition_keys=asset_node["partitionKeys"],
            additional_info=_build_additional_info_for_type(asset_node, partitioning_type),
        )

    @property
    def time_window_partitioning_info(self) -> TimeWindowPartitioningInformation:
        if self.partitioning_type != PartitionDefinitionType.TIME_WINDOW:
            raise Exception(
                f"Partitioning type is {self.partitioning_type}, but expected {PartitionDefinitionType.TIME_WINDOW}"
            )
        if self.additional_info is None:
            raise Exception(
                f"Partitioning type is {self.partitioning_type}, but no additional info was provided."
            )
        return self.additional_info


def _build_additional_info_for_type(
    asset_node: Mapping[str, Any], partitioning_type: PartitionDefinitionType
) -> Optional[TimeWindowPartitioningInformation]:
    if partitioning_type != PartitionDefinitionType.TIME_WINDOW:
        return None
    return TimeWindowPartitioningInformation(fmt=asset_node["partitionDefinition"]["fmt"])


def all_sets_equal(list_of_sets):
    if not list_of_sets:
        return True
    return len(set.union(*list_of_sets)) == len(set.intersection(*list_of_sets))


def translate_logical_date_to_partition_key(
    logical_date: datetime, partitioning_info: PartitioningInformation
) -> str:
    if not partitioning_info.partitioning_type == PartitionDefinitionType.TIME_WINDOW:
        raise Exception(
            "Only time-window partitioned assets or non-partitioned assets are supported out of the box."
        )
    fmt = partitioning_info.time_window_partitioning_info.fmt
    partitions_and_datetimes = [
        (_get_partition_datetime(partition_key, fmt), partition_key)
        for partition_key in partitioning_info.partition_keys
    ]
    matching_partition = next(
        (
            partition_key
            for datetime, partition_key in partitions_and_datetimes
            if datetime.timestamp() == logical_date.timestamp()
        ),
        None,
    )
    if matching_partition is None:
        raise Exception(f"No partition key found for logical date {logical_date}")
    return matching_partition


def _asset_is_partitioned(asset_node: Mapping[str, Any]) -> bool:
    return asset_node["isPartitioned"]


def _get_partition_datetime(partition_key: str, fmt: str) -> datetime:
    try:
        return _add_default_utc_timezone_if_none(datetime.strptime(partition_key, fmt))
    except ValueError:
        raise Exception(f"Could not parse partition key {partition_key} with format {fmt}.")


def _add_default_utc_timezone_if_none(dt: datetime) -> datetime:
    return dt.replace(tzinfo=tz.utc) if dt.tzinfo is None else dt



import re
from collections.abc import Iterable, Mapping, Sequence
from typing import Any, Union

from dagster_airlift.in_airflow.base_asset_operator import BaseDagsterAssetsOperator, Context

UNESCAPED_SLASH_RE = re.compile(r"(?<!\\)/")
ESCAPED_SLASH = "\\/"


class BaseMaterializeAssetsOperator(BaseDagsterAssetsOperator):
    """An operator base class that proxies execution to a user-provided list of Dagster assets.
    Will throw an error at runtime if not all assets can be found on the corresponding Dagster instance.

    Args:
        asset_key_paths (Sequence[Union[str, Sequence[str]]]): A sequence of asset key paths to materialize.
            Each path in the sequence can be a string, which is treated as an asset key path with a single
            component per "/" key, or a sequence of strings representing a path with multiple components. For more,
            see the docs on asset keys: https://docs.dagster.io/guides/build/assets#multi-part-asset-keys
    """

    def __init__(self, asset_key_paths: Sequence[Union[str, Sequence[str]]], *args, **kwargs):
        self.asset_key_paths = [
            _get_path_from_str(path) if isinstance(path, str) else tuple(path)
            for path in asset_key_paths
        ]
        super().__init__(*args, **kwargs)

    def filter_asset_nodes(
        self, context: Context, asset_nodes: Sequence[Mapping[str, Any]]
    ) -> Iterable[Mapping[str, Any]]:
        hashable_path_to_node = {tuple(node["assetKey"]["path"]): node for node in asset_nodes}
        if not all(path in hashable_path_to_node for path in self.asset_key_paths):
            raise ValueError(
                f"Could not find all asset key paths {self.asset_key_paths} in the asset nodes. Found: {list(hashable_path_to_node.keys())}"
            )
        yield from [hashable_path_to_node[path] for path in self.asset_key_paths]


def _get_path_from_str(path_str: str) -> tuple:
    parts = re.split(UNESCAPED_SLASH_RE, path_str)
    return tuple(part.replace(ESCAPED_SLASH, "/") for part in parts)



from collections.abc import Iterable, Iterator, Sequence
from datetime import timedelta
from typing import Optional

from dagster import (
    AssetCheckKey,
    AssetKey,
    AssetMaterialization,
    DefaultSensorStatus,
    RunRequest,
    SensorDefinition,
    SensorEvaluationContext,
    SensorResult,
    _check as check,
    sensor,
)
from dagster._annotations import beta
from dagster._core.definitions.asset_checks.asset_check_evaluation import AssetCheckEvaluation
from dagster._core.definitions.asset_selection import AssetSelection
from dagster._core.definitions.events import AssetObservation
from dagster._core.definitions.repository_definition.repository_definition import (
    RepositoryDefinition,
)
from dagster._core.errors import (
    DagsterInvariantViolationError,
    DagsterUserCodeExecutionError,
    user_code_error_boundary,
)
from dagster._core.storage.dagster_run import DagsterRun, RunsFilter
from dagster._grpc.client import DEFAULT_SENSOR_GRPC_TIMEOUT
from dagster._record import record
from dagster._serdes import deserialize_value, serialize_value
from dagster._time import datetime_from_timestamp, get_current_datetime
from dagster_shared.serdes import whitelist_for_serdes

from dagster_airlift.constants import (
    AUTOMAPPED_TASK_METADATA_KEY,
    DAG_RUN_ID_TAG_KEY,
    EFFECTIVE_TIMESTAMP_METADATA_KEY,
    TASK_ID_TAG_KEY,
)
from dagster_airlift.core.airflow_defs_data import AirflowDefinitionsData, MappedAsset
from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.runtime_representations import DagRun, TaskInstance
from dagster_airlift.core.sensor.event_translation import (
    AssetEvent,
    DagsterEventTransformerFn,
    default_event_transformer,
    get_timestamp_from_materialization,
    synthetic_mats_for_mapped_asset_keys,
    synthetic_mats_for_mapped_dag_asset_keys,
    synthetic_mats_for_peered_dag_asset_keys,
    synthetic_mats_for_task_instance,
)

MAIN_LOOP_TIMEOUT_SECONDS = DEFAULT_SENSOR_GRPC_TIMEOUT - 20
DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS = 30
START_LOOKBACK_SECONDS = 60  # Lookback one minute in time for the initial setting of the cursor.


@whitelist_for_serdes
@record
class AirflowPollingSensorCursor:
    """A cursor that stores the last effective timestamp and the last polled dag id."""

    end_date_gte: Optional[float] = None
    end_date_lte: Optional[float] = None
    dag_query_offset: Optional[int] = None


class AirliftSensorEventTransformerError(DagsterUserCodeExecutionError):
    """Error raised when an error occurs in the event transformer function."""


def check_keys_for_asset_keys(
    repository_def: RepositoryDefinition, asset_keys: set[AssetKey]
) -> Iterable[AssetCheckKey]:
    for assets_def in repository_def.asset_graph.assets_defs:
        for check_spec in assets_def.check_specs:
            if check_spec.asset_key in asset_keys:
                yield check_spec.key


@beta
def build_airflow_polling_sensor(
    *,
    mapped_assets: Sequence[MappedAsset],
    airflow_instance: AirflowInstance,
    event_transformer_fn: DagsterEventTransformerFn = default_event_transformer,
    minimum_interval_seconds: int = DEFAULT_AIRFLOW_SENSOR_INTERVAL_SECONDS,
    default_sensor_status: Optional[DefaultSensorStatus] = None,
) -> SensorDefinition:
    """The constructed sensor polls the Airflow instance for activity, and inserts asset events into Dagster's event log.

    The sensor decides which Airflow dags and tasks to monitor by inspecting the metadata of the passed-in Definitions object `mapped_defs`.
    The metadata performing this mapping is typically set by calls to `assets_with_dag_mappings` and `assets_with_task_mappings`.

    Using the `event_transformer_fn` argument, users can provide a function that transforms the materializations emitted by the sensor.
    The expected return type of this function is an iterable of `AssetMaterialization`, `AssetObservation`, or `AssetCheckEvaluation` objects.
    Each object is expected to have a metadata key `dagster_airlift.constants.EFFECTIVE_TIMESTAMP_METADATA_KEY` which is a `dagster.TimestampMetadataValue` set.
    This allows Dagster to correctly order the materializations in the event stream.

    Args:
        mapped_defs (Definitions): The `Definitions` object containing assets with metadata mapping them to Airflow dags and tasks.
        airflow_instance (AirflowInstance): The Airflow instance to poll for dag runs.
        event_transformer_fn (Optional[DagsterEventTransformerFn]): A function that transforms the materializations emitted by the sensor.
        minimum_interval_seconds (int): The minimum interval in seconds between sensor runs. Defaults to 1.

    Returns:
        Definitions: A `Definitions` object containing the constructed sensor.
    """

    @sensor(
        name=f"{airflow_instance.name}__airflow_dag_status_sensor",
        minimum_interval_seconds=minimum_interval_seconds,
        default_status=default_sensor_status or DefaultSensorStatus.RUNNING,
        # This sensor will only ever execute asset checks and not asset materializations.
        asset_selection=AssetSelection.all_asset_checks(),
    )
    def airflow_dag_sensor(context: SensorEvaluationContext) -> SensorResult:
        """Sensor to report materialization events for each asset as new runs come in."""
        context.log.info(f"************Running sensor for {airflow_instance.name}***********")
        airflow_data = AirflowDefinitionsData(
            airflow_instance=airflow_instance,
            resolved_repository=check.not_none(context.repository_def),
        )
        try:
            cursor = (
                deserialize_value(context.cursor, AirflowPollingSensorCursor)
                if context.cursor
                else AirflowPollingSensorCursor()
            )
        except Exception as e:
            context.log.info(f"Failed to interpret cursor. Starting from scratch. Error: {e}")
            cursor = AirflowPollingSensorCursor()
        current_date = get_current_datetime()
        current_dag_offset = cursor.dag_query_offset or 0
        end_date_gte = (
            cursor.end_date_gte
            or (current_date - timedelta(seconds=START_LOOKBACK_SECONDS)).timestamp()
        )
        end_date_lte = cursor.end_date_lte or current_date.timestamp()
        sensor_iter = batch_iter(
            context=context,
            end_date_gte=end_date_gte,
            end_date_lte=end_date_lte,
            offset=current_dag_offset,
            airflow_data=airflow_data,
        )
        all_asset_events: list[AssetMaterialization] = []
        all_check_keys: set[AssetCheckKey] = set()
        latest_offset = current_dag_offset
        repository_def = check.not_none(context.repository_def)
        while get_current_datetime() - current_date < timedelta(seconds=MAIN_LOOP_TIMEOUT_SECONDS):
            batch_result = next(sensor_iter, None)
            if batch_result is None:
                context.log.info("Received no batch result. Breaking.")
                break
            all_asset_events.extend(batch_result.asset_events)

            all_check_keys.update(
                check_keys_for_asset_keys(repository_def, batch_result.all_asset_keys_materialized)
            )
            latest_offset = batch_result.idx

        if batch_result is not None:  # pyright: ignore[reportPossiblyUnboundVariable]
            new_cursor = AirflowPollingSensorCursor(
                end_date_gte=end_date_gte,
                end_date_lte=end_date_lte,
                dag_query_offset=latest_offset + 1,
            )
        else:
            # We have completed iteration for this range
            new_cursor = AirflowPollingSensorCursor(
                end_date_gte=end_date_lte,
                end_date_lte=None,
                dag_query_offset=0,
            )
        updated_asset_events = _get_transformer_result(
            event_transformer_fn=event_transformer_fn,
            context=context,
            airflow_data=airflow_data,
            all_asset_events=all_asset_events,
        )

        context.update_cursor(serialize_value(new_cursor))

        context.log.info(
            f"************Exiting sensor for {airflow_data.airflow_instance.name}***********"
        )
        return SensorResult(
            asset_events=sorted_asset_events(updated_asset_events, repository_def),
            run_requests=[RunRequest(asset_check_keys=list(all_check_keys))]
            if all_check_keys
            else None,
        )

    return airflow_dag_sensor


def sorted_asset_events(
    asset_events: Sequence[AssetEvent],
    repository_def: RepositoryDefinition,
) -> list[AssetEvent]:
    """Sort materializations by end date and toposort order."""
    topo_aks = repository_def.asset_graph.toposorted_asset_keys
    materializations_and_timestamps = [
        (get_timestamp_from_materialization(mat), mat) for mat in asset_events
    ]
    return [
        sorted_event[1]
        for sorted_event in sorted(
            materializations_and_timestamps, key=lambda x: (x[0], topo_aks.index(x[1].asset_key))
        )
    ]


def _get_transformer_result(
    event_transformer_fn: Optional[DagsterEventTransformerFn],
    context: SensorEvaluationContext,
    airflow_data: AirflowDefinitionsData,
    all_asset_events: Sequence[AssetMaterialization],
) -> Sequence[AssetEvent]:
    if not event_transformer_fn:
        return all_asset_events

    with user_code_error_boundary(
        AirliftSensorEventTransformerError,
        lambda: f"Error occurred during event transformation for {airflow_data.airflow_instance.name}",
    ):
        updated_asset_events = list(event_transformer_fn(context, airflow_data, all_asset_events))

    for asset_event in updated_asset_events:
        if not isinstance(
            asset_event, (AssetMaterialization, AssetObservation, AssetCheckEvaluation)
        ):
            raise DagsterInvariantViolationError(
                f"Event transformer function must return AssetMaterialization, AssetObservation, or AssetCheckEvaluation objects. Got {type(asset_event)}."
            )
        if EFFECTIVE_TIMESTAMP_METADATA_KEY not in asset_event.metadata:
            raise DagsterInvariantViolationError(
                f"All returned events must have an effective timestamp, but {asset_event} does not. An effective timestamp can be used by setting dagster_airlift.constants.EFFECTIVE_TIMESTAMP_METADATA_KEY with a dagster.TimestampMetadataValue."
            )
    return updated_asset_events


@record
class BatchResult:
    idx: int
    asset_events: Sequence[AssetMaterialization]
    all_asset_keys_materialized: set[AssetKey]


def batch_iter(
    context: SensorEvaluationContext,
    end_date_gte: float,
    end_date_lte: float,
    offset: int,
    airflow_data: AirflowDefinitionsData,
) -> Iterator[Optional[BatchResult]]:
    total_processed_runs = 0
    total_entries = 0
    while True:
        latest_offset = total_processed_runs + offset
        runs, total_entries = airflow_data.airflow_instance.get_dag_runs_batch(
            dag_ids=list(airflow_data.dag_ids_with_mapped_asset_keys),
            end_date_gte=datetime_from_timestamp(end_date_gte),
            end_date_lte=datetime_from_timestamp(end_date_lte),
            offset=latest_offset,
        )
        if len(runs) == 0:
            yield None
            context.log.info("Received no runs. Breaking.")
            break
        context.log.info(
            f"Processing {len(runs)}/{total_entries} dag runs for {airflow_data.airflow_instance.name}..."
        )
        for i, dag_run in enumerate(runs):
            context.log.info(
                f"dag ids: {[dag_id for dag_id in airflow_data.dag_ids_with_mapped_asset_keys]}"
            )
            mats = build_synthetic_asset_materializations(
                context, airflow_data.airflow_instance, dag_run, airflow_data
            )
            context.log.info(f"Found {len(mats)} materializations for {dag_run.run_id}")

            all_asset_keys_materialized = {mat.asset_key for mat in mats}
            yield (
                BatchResult(
                    idx=i + latest_offset,
                    asset_events=mats,
                    all_asset_keys_materialized=all_asset_keys_materialized,
                )
                if mats
                else None
            )
        total_processed_runs += len(runs)
        context.log.info(
            f"Processed {total_processed_runs}/{total_entries} dag runs for {airflow_data.airflow_instance.name}."
        )
        if total_processed_runs == total_entries:
            yield None
            context.log.info("Processed all runs. Breaking.")
            break


def build_synthetic_asset_materializations(
    context: SensorEvaluationContext,
    airflow_instance: AirflowInstance,
    dag_run: DagRun,
    airflow_data: AirflowDefinitionsData,
) -> list[AssetMaterialization]:
    """In this function we need to return the asset materializations we want to synthesize
    on behalf of the user.

    This happens when the user has modeled an external asset that has a corresponding
    task in airflow which is not proxied to Dagster. We want to detect the case
    where there is a successful airflow task instance that is mapped to a
    dagster asset but is _not_ proxied, and then synthensize a materialization
    for observability.

    We do this by querying for successful task instances in Airflow. And then
    for each successful task we see it there exists a Dagster Run tagged with
    the run id. If there is not Dagster run, we know the task was not proxied.

    Task instances are mutable in Airflow, so we are not guaranteed to register
    every task instance. If, for example, the sensor is paused, and then there are
    multiple task clearings, we will only register the last materialization.

    This also currently does not support dynamic tasks in Airflow, in which case
    the use should instead map at the dag-level granularity.
    """
    # https://linear.app/dagster-labs/issue/FOU-444/make-sensor-work-with-an-airflow-dag-run-that-has-more-than-1000
    dagster_runs = context.instance.get_runs(
        filters=RunsFilter(tags={DAG_RUN_ID_TAG_KEY: dag_run.run_id}),
        limit=1000,
    )
    context.log.info(f"Found {len(dagster_runs)} dagster runs for {dag_run.run_id}")

    context.log.info(
        f"Airlift Sensor: Found dagster run ids: {[run.run_id for run in dagster_runs]}"
        f" for airflow run id {dag_run.run_id} and dag id {dag_run.dag_id}"
    )
    synthetic_mats = []
    # Peered dag-level materializations will always be emitted.
    synthetic_mats.extend(synthetic_mats_for_peered_dag_asset_keys(dag_run, airflow_data))
    # If there is a dagster run for this dag, we don't need to synthesize materializations for mapped dag assets.
    if not dagster_runs:
        synthetic_mats.extend(synthetic_mats_for_mapped_dag_asset_keys(dag_run, airflow_data))
    synthetic_mats.extend(
        get_synthetic_task_mats(
            airflow_instance=airflow_instance,
            dagster_runs=dagster_runs,
            dag_run=dag_run,
            airflow_data=airflow_data,
            context=context,
        )
    )
    return synthetic_mats


def get_synthetic_task_mats(
    airflow_instance: AirflowInstance,
    dagster_runs: Sequence[DagsterRun],
    dag_run: DagRun,
    airflow_data: AirflowDefinitionsData,
    context: SensorEvaluationContext,
) -> list[AssetMaterialization]:
    task_instances = airflow_instance.get_task_instance_batch(
        run_id=dag_run.run_id,
        dag_id=dag_run.dag_id,
        task_ids=[task_id for task_id in airflow_data.task_ids_in_dag(dag_run.dag_id)],
        states=["success"],
    )
    check.invariant(
        len({ti.task_id for ti in task_instances}) == len(task_instances),
        "Assuming one task instance per task_id for now. Dynamic Airflow tasks not supported.",
    )
    synthetic_mats = []
    context.log.info(f"Found {len(task_instances)} task instances for {dag_run.run_id}")
    context.log.info(f"All task instances {task_instances}")
    dagster_runs_by_task_id = {
        run.tags[TASK_ID_TAG_KEY]: run for run in dagster_runs if TASK_ID_TAG_KEY in run.tags
    }
    task_instances_by_task_id = {ti.task_id: ti for ti in task_instances}
    for task_id, task_instance in task_instances_by_task_id.items():
        # No dagster runs means that the computation that materializes the asset was not proxied to Dagster.
        # Therefore the dags ran completely in Airflow, and we will synthesize materializations in Dagster corresponding to that data run.
        if task_id not in dagster_runs_by_task_id:
            context.log.info(
                f"Synthesizing materialization for tasks {task_id} in dag {dag_run.dag_id} because no dagster run found."
            )
            synthetic_mats.extend(
                synthetic_mats_for_task_instance(airflow_data, dag_run, task_instance)
            )
        else:
            # We *always* emit for the automapped tasks, even if they are proxied
            asset_keys_to_emit = automapped_tasks_asset_keys(dag_run, airflow_data, task_instance)

            synthetic_mats.extend(
                synthetic_mats_for_mapped_asset_keys(
                    dag_run=dag_run, task_instance=task_instance, asset_keys=asset_keys_to_emit
                )
            )

            context.log.info(
                f"Dagster run found for task {task_id} in dag {dag_run.dag_id}. Run {dagster_runs_by_task_id[task_id].run_id}"
            )
    return synthetic_mats


def automapped_tasks_asset_keys(
    dag_run: DagRun, airflow_data: AirflowDefinitionsData, task_instance: TaskInstance
) -> set[AssetKey]:
    asset_keys_to_emit = set()
    asset_keys = airflow_data.asset_keys_in_task(dag_run.dag_id, task_instance.task_id)
    for asset_key in asset_keys:
        spec = airflow_data.airflow_mapped_asset_specs[asset_key]
        if spec.metadata.get(AUTOMAPPED_TASK_METADATA_KEY):
            asset_keys_to_emit.add(asset_key)
    return asset_keys_to_emit



from collections.abc import Iterator, Sequence
from dataclasses import dataclass, field
from pathlib import Path
from typing import Annotated, Any, Literal, Optional, TypeAlias, Union

from dagster import ComponentLoadContext, Resolvable
from dagster._core.definitions.asset_key import AssetKey
from dagster._core.definitions.assets.definition.asset_spec import (
    SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET,
    AssetSpec,
)
from dagster._core.definitions.definitions_class import Definitions
from dagster.components.component.state_backed_component import StateBackedComponent
from dagster.components.component_scaffolding import scaffold_component
from dagster.components.core.defs_module import DefsFolderComponent, find_components_from_context
from dagster.components.resolved.base import resolve_fields
from dagster.components.resolved.context import ResolutionContext
from dagster.components.resolved.core_models import ResolvedAssetKey, ResolvedAssetSpec
from dagster.components.resolved.model import Resolver
from dagster.components.scaffold.scaffold import Scaffolder, ScaffoldRequest, scaffold_with
from dagster.components.utils.defs_state import (
    DefsStateConfig,
    DefsStateConfigArgs,
    ResolvedDefsStateConfig,
)
from dagster_shared.serdes.serdes import deserialize_value, serialize_value
from pydantic import BaseModel

import dagster_airlift.core as dg_airlift_core
from dagster_airlift.core.airflow_instance import AirflowAuthBackend
from dagster_airlift.core.basic_auth import AirflowBasicAuthBackend
from dagster_airlift.core.filter import AirflowFilter
from dagster_airlift.core.load_defs import (
    _apply_airflow_data_to_specs,
    _get_dag_to_spec_mapping,
    build_airflow_monitoring_defs,
    construct_dag_jobs,
    construct_dataset_specs,
    replace_assets_in_defs,
    type_narrow_defs_assets,
)
from dagster_airlift.core.serialization.compute import compute_serialized_data
from dagster_airlift.core.serialization.serialized_data import (
    DagHandle,
    SerializedAirflowDefinitionsData,
    TaskHandle,
)


@dataclass
class ResolvedAirflowBasicAuthBackend(Resolvable):
    type: Literal["basic_auth"]
    webserver_url: str
    username: str
    password: str


@dataclass
class ResolvedAirflowMwaaAuthBackend(Resolvable):
    type: Literal["mwaa"]
    env_name: str
    region_name: Optional[str] = None
    profile_name: Optional[str] = None
    aws_account_id: Optional[str] = None
    aws_access_key_id: Optional[str] = None
    aws_secret_access_key: Optional[str] = None
    aws_session_token: Optional[str] = None


@dataclass
class InAirflowAsset(Resolvable):
    spec: ResolvedAssetSpec


@dataclass
class InDagsterAssetRef(Resolvable):
    by_key: ResolvedAssetKey


def resolve_mapped_asset(context: ResolutionContext, model) -> Union[AssetKey, AssetSpec]:
    if isinstance(model, InAirflowAsset.model()):
        return InAirflowAsset.resolve_from_model(context, model).spec
    elif isinstance(model, InDagsterAssetRef.model()):
        return InDagsterAssetRef.resolve_from_model(context, model).by_key
    else:
        raise ValueError(f"Unsupported asset type: {type(model)}")


ResolvedMappedAsset: TypeAlias = Annotated[
    Union[AssetKey, AssetSpec],
    Resolver(
        resolve_mapped_asset,
        model_field_type=Union[InAirflowAsset.model(), InDagsterAssetRef.model()],
    ),
]


@dataclass
class AirflowFilterParams(Resolvable):
    dag_id_ilike: Optional[str] = None
    airflow_tags: Optional[Sequence[str]] = None
    retrieve_datasets: bool = True
    dataset_uri_ilike: Optional[str] = None


def resolve_airflow_filter(context: ResolutionContext, model) -> AirflowFilter:
    return AirflowFilter(**resolve_fields(model, AirflowFilterParams, context))


ResolvedAirflowFilter: TypeAlias = Annotated[
    AirflowFilter,
    Resolver(
        resolve_airflow_filter,
        model_field_type=AirflowFilterParams.model(),
    ),
]


@dataclass
class AirflowTaskMapping(Resolvable):
    task_id: str
    assets: Sequence[ResolvedMappedAsset]


@dataclass
class AirflowDagMapping(Resolvable):
    dag_id: str
    assets: Optional[Sequence[ResolvedMappedAsset]] = None
    task_mappings: Optional[Sequence[AirflowTaskMapping]] = None


class AirflowInstanceScaffolderParams(BaseModel):
    name: str
    auth_type: Literal["basic_auth", "mwaa"]


class AirflowInstanceScaffolder(Scaffolder[AirflowInstanceScaffolderParams]):
    @classmethod
    def get_scaffold_params(cls) -> type[AirflowInstanceScaffolderParams]:
        return AirflowInstanceScaffolderParams

    def scaffold(self, request: ScaffoldRequest[AirflowInstanceScaffolderParams]) -> None:
        full_params: dict[str, Any] = {
            "name": request.params.name,
        }
        if request.params.auth_type == "basic_auth":
            full_params["auth"] = {
                "type": "basic_auth",
                "webserver_url": '{{ env("AIRFLOW_WEBSERVER_URL") }}',
                "username": '{{ env("AIRFLOW_USERNAME") }}',
                "password": '{{ env("AIRFLOW_PASSWORD") }}',
            }
        else:
            raise ValueError(f"Unsupported auth type: {request.params.auth_type}")
        scaffold_component(request, full_params)


def resolve_auth(context: ResolutionContext, auth) -> AirflowAuthBackend:
    if auth.type == "basic_auth":
        resolved_model_fields = resolve_fields(auth, ResolvedAirflowBasicAuthBackend, context)
        return AirflowBasicAuthBackend(
            webserver_url=resolved_model_fields["webserver_url"],
            username=resolved_model_fields["username"],
            password=resolved_model_fields["password"],
        )
    elif auth.type == "mwaa":
        resolved_model_fields = resolve_fields(auth, ResolvedAirflowMwaaAuthBackend, context)
        try:
            from dagster_airlift.mwaa import MwaaSessionAuthBackend
        except ImportError:
            raise ValueError(
                "When resolving an AirflowInstance with MWAA auth, dagster-airlift.mwaa was not installed. Please install the `dagster-airlift[mwaa]` subpackage."
            )
        try:
            import boto3
        except ImportError:
            raise ValueError(
                "When resolving an AirflowInstance with MWAA auth, boto3 was not installed. Please ensure boto3 is installed in your environment."
            )
        boto_session = boto3.Session(
            region_name=resolved_model_fields.get("region_name"),
            profile_name=resolved_model_fields.get("profile_name"),
            aws_access_key_id=resolved_model_fields.get("aws_access_key_id"),
            aws_secret_access_key=resolved_model_fields.get("aws_secret_access_key"),
            aws_session_token=resolved_model_fields.get("aws_session_token"),
            aws_account_id=resolved_model_fields.get("aws_account_id"),
        )
        return MwaaSessionAuthBackend(
            mwaa_client=boto_session.client("mwaa"),
            env_name=resolved_model_fields["env_name"],
        )
    else:
        raise ValueError(f"Unsupported auth type: {auth.type}")


ResolvedAirflowAuthBackend: TypeAlias = Annotated[
    AirflowAuthBackend,
    Resolver(
        resolve_auth,
        model_field_type=Union[
            ResolvedAirflowBasicAuthBackend.model(), ResolvedAirflowMwaaAuthBackend.model()
        ],
    ),
]


@scaffold_with(AirflowInstanceScaffolder)
@dataclass
class AirflowInstanceComponent(StateBackedComponent, Resolvable):
    """Loads Airflow DAGs and tasks from an Airflow instance as Dagster assets.

    This component connects to an Airflow instance, retrieves metadata about DAGs and tasks,
    and creates corresponding Dagster assets. It supports mapping Airflow tasks to existing
    Dagster assets or creating new assets to represent Airflow workflows.

    Example:

        .. code-block:: yaml

            # defs.yaml

            type: dagster_airlift.core.AirflowInstanceComponent
            attributes:
              name: my_airflow_instance
              auth:
                type: basic_auth
                webserver_url: "{{ env.AIRFLOW_WEBSERVER_URL }}"
                username: "{{ env.AIRFLOW_USERNAME }}"
                password: "{{ env.AIRFLOW_PASSWORD }}"
              filter:
                dag_id_ilike: "analytics_%"
                retrieve_datasets: true
    """

    auth: ResolvedAirflowAuthBackend
    name: str
    filter: Optional[ResolvedAirflowFilter] = None
    mappings: Optional[Sequence[AirflowDagMapping]] = None
    source_code_retrieval_enabled: Optional[bool] = None
    defs_state: ResolvedDefsStateConfig = field(
        default_factory=DefsStateConfigArgs.legacy_code_server_snapshots
    )

    @property
    def defs_state_config(self) -> DefsStateConfig:
        default_key = f"{self.__class__.__name__}[{self.name}]"
        return DefsStateConfig.from_args(self.defs_state, default_key=default_key)

    def _get_instance(self) -> dg_airlift_core.AirflowInstance:
        return dg_airlift_core.AirflowInstance(
            auth_backend=self.auth,
            name=self.name,
        )

    async def write_state_to_path(self, state_path: Path) -> None:
        # Fetch the serialized Airflow definitions data
        state = compute_serialized_data(
            airflow_instance=self._get_instance(),
            mapped_assets=[],
            dag_selector_fn=None,
            automapping_enabled=False,
            source_code_retrieval_enabled=self.source_code_retrieval_enabled,
            retrieval_filter=self.filter or AirflowFilter(),
        )
        state_path.write_text(serialize_value(state))

    def build_defs_from_state(
        self, context: ComponentLoadContext, state_path: Optional[Path]
    ) -> Definitions:
        if state_path is None:
            return Definitions()

        # Load the serialized state
        serialized_airflow_data = deserialize_value(
            state_path.read_text(), SerializedAirflowDefinitionsData
        )

        # Get mapped defs from subdirs
        mapped_defs = apply_mappings(defs_from_subdirs(context), self.mappings or [])
        mapped_assets = type_narrow_defs_assets(mapped_defs)

        # Apply airflow data to specs
        assets_with_airflow_data = _apply_airflow_data_to_specs(
            [
                *mapped_assets,
                *construct_dataset_specs(serialized_airflow_data),
            ],
            serialized_airflow_data,
        )

        # Construct DAG jobs
        dag_to_spec_mapping = _get_dag_to_spec_mapping(assets_with_airflow_data)
        jobs = construct_dag_jobs(
            serialized_data=serialized_airflow_data,
            mapped_specs=dag_to_spec_mapping,
        )

        # Build the final definitions
        return Definitions.merge(
            replace_assets_in_defs(defs=mapped_defs, assets=assets_with_airflow_data),
            Definitions(jobs=jobs),
            build_airflow_monitoring_defs(airflow_instance=self._get_instance()),
        )


def defs_from_subdirs(context: ComponentLoadContext) -> Definitions:
    """Get all definitions coming from subdirectories in the current context. We explicitly don't want to include the current directory in the search, to avoid including the current component."""
    return DefsFolderComponent(
        path=context.path,
        children=find_components_from_context(context),
    ).build_defs(context)


def handle_iterator(
    mappings: Optional[Sequence[AirflowDagMapping]],
) -> Iterator[tuple[Union[TaskHandle, DagHandle], Sequence[ResolvedMappedAsset]]]:
    if mappings is None:
        return
    for mapping in mappings:
        for task_mapping in mapping.task_mappings or []:
            yield (
                TaskHandle(dag_id=mapping.dag_id, task_id=task_mapping.task_id),
                task_mapping.assets,
            )
        yield DagHandle(dag_id=mapping.dag_id), mapping.assets or []


def apply_mappings(defs: Definitions, mappings: Sequence[AirflowDagMapping]) -> Definitions:
    specs = {spec.key: spec for spec in defs.resolve_all_asset_specs()}
    key_to_handle_mapping = {}
    additional_assets = []

    for handle, asset_refs in handle_iterator(mappings):
        if not asset_refs:
            continue
        for asset in asset_refs:
            if isinstance(asset, AssetKey):
                if asset not in specs:
                    raise ValueError(f"Asset with key {asset} not found in definitions")
                key_to_handle_mapping[asset] = handle
            elif isinstance(asset, AssetSpec):
                if asset.key in specs and not is_autocreated_stub_asset(specs[asset.key]):
                    raise ValueError(f"Asset with key {asset.key} already exists in definitions")
                additional_assets.append(
                    asset.merge_attributes(metadata={handle.metadata_key: [handle.as_dict]})
                )

    def spec_mapper(spec: AssetSpec) -> AssetSpec:
        if spec.key in key_to_handle_mapping:
            handle = key_to_handle_mapping[spec.key]
            return spec.merge_attributes(metadata={handle.metadata_key: [handle.as_dict]})
        return spec

    return Definitions.merge_unbound_defs(
        defs.map_resolved_asset_specs(func=spec_mapper), Definitions(assets=additional_assets)
    )


def is_autocreated_stub_asset(spec: AssetSpec) -> bool:
    return spec.metadata.get(SYSTEM_METADATA_KEY_AUTO_CREATED_STUB_ASSET) is True



from pathlib import Path

from setuptools import find_packages, setup


def get_version() -> str:
    version: dict[str, str] = {}
    with open(Path(__file__).parent / "dagster_airlift/version.py", encoding="utf8") as fp:
        exec(fp.read(), version)

    return version["__version__"]


ver = get_version()
pin = "" if ver == "1!0+dev" or "rc" in ver else f"=={ver}"

# The [in-airflow] subpackage does not have a setup dependency on Airflow because
# Airflow cannot be installed via setup.py reliably. Instead, users need to install
# from a constraints file as recommended by the Airflow project.
# However, to ensure a reliable test and tutorial setup, we pin a version of Airflow
# that is compatible with the current version of dagster-airlift for all supported
# versions of python.
# Eventually, we could consider adding a test suite that runs across different versions of airflow
# to ensure compatibility.
AIRFLOW_REQUIREMENTS = [
    # Requirements for python versions under 3.12.
    "apache-airflow==2.9.3; python_version < '3.12'",
    "marshmallow==3.20.1; python_version < '3.12'",
    "pendulum>=2.0.0,<4.0.0; python_version < '3.12'",
    # Requirements for python versions 3.12 and above.
    "apache-airflow>=2.9.0; python_version >= '3.12'",
    "marshmallow==3.23.1; python_version >= '3.12'",
    "pendulum>=3.0.0; python_version >= '3.12'",
    # Flask-session 0.6 is incompatible with certain airflow-provided test
    # utilities.
    "flask-session<0.6.0",
    # https://github.com/apache/airflow/issues/35234
    "connexion<3.0.0",
    # https://github.com/apache/airflow/discussions/57769#discussioncomment-14861217
    "fastapi<0.118",
]

CLI_REQUIREMENTS = ["click", "structlog"]


setup(
    name="dagster-airlift",
    version=ver,
    author="Dagster Labs",
    author_email="hello@dagsterlabs.com",
    license="Apache-2.0",
    description="A toolkit for observing integration and migration between Apache Airflow and Dagster.",
    url=(
        "https://github.com/dagster-io/dagster/tree/master/python_modules/libraries/dagster-airlift"
    ),
    classifiers=[
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: OS Independent",
    ],
    packages=find_packages(exclude=["dagster_airlift_tests*", "kitchen-sink", "perf-harness"]),
    requires=CLI_REQUIREMENTS,
    extras_require={
        "core": [
            "dagster==1.12.6",
            *CLI_REQUIREMENTS,
        ],
        # [in-airflow] doesn't directly have a dependency on airflow because Airflow cannot be installed via setup.py reliably. Instead, users need to install from a constraints
        # file as recommended by the Airflow project.
        "in-airflow": CLI_REQUIREMENTS,
        # [tutorial] includes additional dependencies needed to run the tutorial. Namely, the dagster-webserver and the constrained airflow packages.
        "tutorial": [
            "dagster-webserver",
            *AIRFLOW_REQUIREMENTS,
            *CLI_REQUIREMENTS,
        ],
        "mwaa": [
            "boto3>=1.18.0"
        ],  # confirms that mwaa is available in the environment (can't find exactly which version adds mwaa support, but I can confirm that 1.18.0 and greater have it.)
        "test": [
            "pytest",
            "dagster-dbt==0.28.6",
            "dbt-duckdb",
            "boto3",
            "dagster-webserver==1.12.6",
            *AIRFLOW_REQUIREMENTS,
            *CLI_REQUIREMENTS,
        ],
    },
    entry_points={
        "console_scripts": [
            "dagster-airlift = dagster_airlift.cli:cli",
        ],
        "dagster_dg_cli.registry_modules": [
            "dagster_airlift.core.components = dagster_airlift.core.components",
        ],
    },
    zip_safe=False,
)



import os
import signal
import subprocess
import sys
import time
from collections.abc import Callable, Generator
from contextlib import contextmanager
from pathlib import Path
from tempfile import TemporaryDirectory
from typing import Any, Optional

import psutil
import pytest
import requests
from dagster._core.test_utils import environ
from dagster._time import get_current_timestamp
from dagster._utils import process_is_alive

from dagster_airlift.core.airflow_instance import AirflowInstance
from dagster_airlift.core.basic_auth import AirflowBasicAuthBackend


####################################################################################################
# AIRFLOW SETUP FIXTURES
# Sets up the airflow environment for testing. Running at localhost:8080.
# Callsites are expected to provide implementations for dags_dir fixture.
####################################################################################################
def _airflow_is_ready(*, port: int, expected_num_dags: int) -> bool:
    try:
        af_instance = AirflowInstance(
            auth_backend=AirflowBasicAuthBackend(
                webserver_url=f"http://localhost:{port}",
                username="admin",
                password="admin",
            ),
            name="test",
        )
        return len(af_instance.list_dags()) >= expected_num_dags
    except:
        return False


@pytest.fixture(name="airflow_home")
def default_airflow_home() -> Generator[str, None, None]:
    with TemporaryDirectory() as tmpdir:
        with environ({"AIRFLOW_HOME": tmpdir}):
            yield tmpdir


@pytest.fixture(name="setup")
def setup_fixture(airflow_home: Path, dags_dir: Path) -> Generator[Path, None, None]:
    assert os.environ["AIRFLOW_HOME"] == str(airflow_home), "AIRFLOW_HOME is not set correctly"
    temp_env = {
        **os.environ.copy(),
        # Provide link to local dagster installation.
        "DAGSTER_URL": "http://localhost:3333",
    }
    path_to_script = Path(__file__).parent.parent.parent / "scripts" / "airflow_setup.sh"
    subprocess.run(["chmod", "+x", path_to_script], check=True, env=temp_env)
    subprocess.run([path_to_script, dags_dir, airflow_home], check=True, env=temp_env)
    with environ(temp_env):
        yield airflow_home


@pytest.fixture(name="reserialize_dags")
def reserialize_fixture(airflow_instance: None) -> Callable[[], None]:
    """Forces airflow to reserialize dags, to ensure that the latest changes are picked up."""

    def _reserialize_dags() -> None:
        subprocess.check_output(["airflow", "dags", "reserialize"])

    return _reserialize_dags


@contextmanager
def stand_up_airflow(
    env: Any = {},
    airflow_cmd: list[str] = ["airflow", "standalone"],
    cwd: Optional[Path] = None,
    stdout_channel: Optional[int] = None,
    port: int = 8080,
    expected_num_dags: int = 1,
) -> Generator[subprocess.Popen, None, None]:
    process = subprocess.Popen(
        airflow_cmd,
        cwd=cwd,
        env=env,  # since we have some temp vars in the env right now
        shell=False,
        preexec_fn=os.setsid,  # noqa  # fuck it we ball
        stdout=stdout_channel,
    )
    try:
        # Give airflow a second to stand up
        time.sleep(5)
        initial_time = get_current_timestamp()

        airflow_ready = False
        while get_current_timestamp() - initial_time < 60:
            if _airflow_is_ready(port=port, expected_num_dags=expected_num_dags):
                airflow_ready = True
                break
            time.sleep(1)

        assert airflow_ready, "Airflow did not start within 60 seconds..."
        yield process
    finally:
        if process_is_alive(process.pid):
            # Kill process group, since process.kill and process.terminate do not work.
            os.killpg(process.pid, signal.SIGKILL)
            process.wait(5)


@pytest.fixture(name="airflow_instance")
def airflow_instance_fixture(setup: None) -> Generator[subprocess.Popen, None, None]:
    with stand_up_airflow(env=os.environ) as process:
        yield process


####################################################################################################
# DAGSTER SETUP FIXTURES
# Sets up the dagster environment for testing. Running at localhost:3333.
# Callsites are expected to provide implementations for dagster_defs_path fixture.
####################################################################################################
def _dagster_is_ready(port: int) -> bool:
    try:
        response = requests.get(f"http://localhost:{port}")
        return response.status_code == 200
    except:
        return False


@pytest.fixture(name="dagster_home")
def setup_dagster_home() -> Generator[str, None, None]:
    """Instantiate a temporary directory to serve as the DAGSTER_HOME."""
    with TemporaryDirectory() as tmpdir:
        with environ({"DAGSTER_HOME": tmpdir}):
            yield tmpdir


@pytest.fixture(name="dagster_dev_cmd")
def dagster_dev_cmd(dagster_defs_path: str) -> list[str]:
    """Return the command used to stand up dagster dev."""
    return ["dagster", "dev", "-f", dagster_defs_path, "-p", "3333"]


@pytest.fixture(name="dagster_dev")
def setup_dagster(
    airflow_instance: None, dagster_home: str, dagster_dev_cmd: list[str]
) -> Generator[Any, None, None]:
    # The version of airflow we use on 3.12 or greater (2.10.2) takes longer to reconcile the dags, and sometimes does it partially.
    # We need to wait for all the dags to be loaded before we can start dagster.
    # A better solution would be to poll the airflow instance for the existence of all expected dags, but this is a stopgap.
    if sys.version_info >= (3, 12):
        time.sleep(20)
    with stand_up_dagster(dagster_dev_cmd) as process:
        yield process
    time.sleep(5)


@contextmanager
def stand_up_dagster(
    dagster_dev_cmd: list[str], port: int = 3333
) -> Generator[subprocess.Popen, None, None]:
    """Stands up a dagster instance using the dagster dev CLI. dagster_defs_path must be provided
    by a fixture included in the callsite.
    """
    process = subprocess.Popen(
        dagster_dev_cmd,
        env=os.environ.copy(),
        shell=False,
        preexec_fn=os.setsid,  # noqa
    )
    try:
        # Give dagster a second to stand up
        time.sleep(5)

        dagster_ready = False
        initial_time = get_current_timestamp()
        while get_current_timestamp() - initial_time < 60:
            if _dagster_is_ready(port):
                dagster_ready = True
                break
            time.sleep(1)

        assert dagster_ready, "Dagster did not start within 30 seconds..."
        yield process
    finally:
        if psutil.Process(pid=process.pid).is_running():
            # Kill process group, since process.kill and process.terminate do not work.
            os.killpg(process.pid, signal.SIGKILL)


####################################################################################################
# MISCELLANEOUS FIXTURES
# Fixtures that are useful across contexts.
####################################################################################################
